"""Acquisition function tests."""

from hypothesis import given

from baybe.recommenders import BotorchRecommender

from .hypothesis_strategies.acquisition import acquisition_functions


@given(acquisition_functions)
def test_acqfs(acqf):
    """Test all acquisition functions with sequential greedy recommender."""
    BotorchRecommender(acquisition_function=acqf)



## Example for using synthetic test functions in hybrid spaces

# This examples shows how to optimize a custom test function in a hybrid searchspace.
# It focuses on the searchspace-related aspects and not on the custom test function.

# This example assumes some basic familiarity with using BayBE and synthetic test functions.
# We thus refer to [`campaign`](./../Basics/campaign.md) for a basic example.
# For details on using synthetic test functions, we refer to other examples in this directory.

### Necessary imports for this example

import numpy as np
import pandas as pd
from botorch.test_functions import Rastrigin

from baybe import Campaign
from baybe.parameters import NumericalContinuousParameter, NumericalDiscreteParameter
from baybe.recommenders import NaiveHybridSpaceRecommender, TwoPhaseMetaRecommender
from baybe.searchspace import SearchSpace
from baybe.targets import NumericalTarget
from baybe.utils.dataframe import arrays_to_dataframes

### Defining the test function and the hybrid dimensions

# See [`discrete_space`](./discrete_space.md) for details on the test function.

DIMENSION = 6

# This examples uses two lists containing the discrete resp. discrete dimensions.
# Together, these should contain the integers `0,1,...,DIMENSION-1`.
# Although this is checked also in this file, you should configure these indices manually
# manually here and verify that the experiment you set up is configured correctly.
# In particular, if the function that you want to use is only available for a fixed
# dimension, then these will be overwritten by distributing the first half of the
# dimension to `DISC_INDICES` and the remaining ones to `CONT_INDICES`.

DISC_INDICES = [0, 1, 2]
CONT_INDICES = [3, 4, 5]

TestFunctionClass = Rastrigin

# This part checks if the test function already has a fixed dimension.
# In that case, we print a warning and replace DIMENSION.

if not hasattr(TestFunctionClass, "dim"):
    TestFunction = TestFunctionClass(dim=DIMENSION)
elif TestFunctionClass().dim == DIMENSION:
    TestFunction = TestFunctionClass()
else:
    print(
        f"\nYou choose a dimension of {DIMENSION} for the test function"
        f"{TestFunctionClass}. However, this function can only be used in "
        f"{TestFunctionClass().dim} dimension, so the provided dimension is replaced. "
        "Also, DISC_INDICES and CONT_INDICES will be re-written."
    )
    TestFunction = TestFunctionClass()
    DIMENSION = TestFunctionClass().dim
    DISC_INDICES = list(range(0, (DIMENSION + 1) // 2))
    CONT_INDICES = list(range((DIMENSION + 1) // 2, DIMENSION))

# This check verifies if the union of the given index sets yields indices matching `DIMENSION`.
# If this fails, then either the intersection between the index sets is not empty or the test
# function has another dimension.
# Note that this might in particular happen for test functions that ignore the `dim` keyword!

if set(CONT_INDICES + DISC_INDICES) != set(range(DIMENSION)):
    raise ValueError(
        "Either the intersection between CONT_IND and DISC_IND is not empty or your "
        "indices do not match."
    )

BOUNDS = TestFunction.bounds

### Constructing the hybrid searchspace

# The following parameter decides how many points each discrete dimension should have.

POINTS_PER_DIM = 3

# Construct the continuous parameters as `NumericContinuous` parameters.

cont_parameters = [
    NumericalContinuousParameter(
        name=f"x_{k+1}",
        bounds=(BOUNDS[0, k], BOUNDS[1, k]),
    )
    for k in CONT_INDICES
]

# Construct the discrete parameters as `NumericalDiscreteParameters`.

disc_parameters = [
    NumericalDiscreteParameter(
        name=f"x_{k+1}",
        values=list(np.linspace(BOUNDS[0, k], BOUNDS[1, k], POINTS_PER_DIM)),
        tolerance=0.01,
    )
    for k in DISC_INDICES
]

searchspace = SearchSpace.from_product(parameters=disc_parameters + cont_parameters)
target = NumericalTarget(name="Target", mode="MIN")
objective = target.to_objective()

### Wrap the test function as a dataframe-based lookup callable

lookup = arrays_to_dataframes(
    searchspace.parameter_names, [target.name], use_torch=True
)(TestFunction)


### Constructing hybrid recommenders

# Here, we explicitly create a recommender object to use the `NaiveHybridSpaceRecommender`.
# The keywords `disc_recommender` and `cont_recommender` can be used to select different
# recommenders for the corresponding subspaces.
# We use the default choices, which is the `BotorchRecommender`.

hybrid_recommender = TwoPhaseMetaRecommender(recommender=NaiveHybridSpaceRecommender())

### Constructing the campaign and performing a recommendation

campaign = Campaign(
    searchspace=searchspace,
    objective=objective,
    recommender=hybrid_recommender,
)

# Get a recommendation for a fixed batch size.
BATCH_SIZE = 3
recommendation = campaign.recommend(batch_size=BATCH_SIZE)

# Evaluate the test function.

lookup_values = lookup(recommendation)
measurements = pd.concat([recommendation, lookup_values], axis=1)

# Here, we inform the campaign about our measurement.

campaign.add_measurements(measurements)
print("\n\nRecommended experiments with measured values: ")
print(measurements)



"""Benchmark domains."""

from benchmarks.definition.config import Benchmark
from benchmarks.domains.synthetic_2C1D_1C import synthetic_2C1D_1C_benchmark

BENCHMARKS: list[Benchmark] = [
    synthetic_2C1D_1C_benchmark,
]

__all__ = ["BENCHMARKS"]



"""Benchmark configurations."""

import time
from abc import ABC
from collections.abc import Callable
from datetime import datetime, timedelta, timezone
from typing import Any, Generic, TypeVar

from attrs import define, field
from attrs.validators import instance_of
from cattr.gen import make_dict_unstructure_fn, override
from pandas import DataFrame

from baybe.utils.random import temporary_seed
from benchmarks.result import Result, ResultMetadata
from benchmarks.serialization import BenchmarkSerialization, converter


@define(frozen=True)
class BenchmarkSettings(ABC, BenchmarkSerialization):
    """Benchmark configuration for recommender analyses."""

    random_seed: int = field(validator=instance_of(int), kw_only=True, default=1337)
    """The random seed for reproducibility."""


BenchmarkSettingsType = TypeVar("BenchmarkSettingsType", bound=BenchmarkSettings)


@define(frozen=True)
class ConvergenceExperimentSettings(BenchmarkSettings):
    """Benchmark configuration for recommender convergence analyses."""

    batch_size: int = field(validator=instance_of(int))
    """The recommendation batch size."""

    n_doe_iterations: int = field(validator=instance_of(int))
    """The number of Design of Experiment iterations."""

    n_mc_iterations: int = field(validator=instance_of(int))
    """The number of Monte Carlo iterations."""


@define(frozen=True)
class Benchmark(Generic[BenchmarkSettingsType], BenchmarkSerialization):
    """The base class for a benchmark executable."""

    settings: BenchmarkSettingsType = field()
    """The benchmark configuration."""

    function: Callable[[BenchmarkSettingsType], DataFrame] = field()
    """The callable which contains the benchmarking logic."""

    name: str = field(init=False)
    """The name of the benchmark."""

    best_possible_result: float | None = field(default=None)
    """The best possible result which can be achieved in the optimization process."""

    optimal_function_inputs: list[dict[str, Any]] | None = field(default=None)
    """An input that creates the best_possible_result."""

    @property
    def description(self) -> str:
        """The description of the benchmark function."""
        if self.function.__doc__ is not None:
            return self.function.__doc__
        return "No description available."

    @name.default
    def _default_name(self):
        """Return the name of the benchmark function."""
        return self.function.__name__

    def __call__(self) -> Result:
        """Execute the benchmark and return the result."""
        start_datetime = datetime.now(timezone.utc)

        with temporary_seed(self.settings.random_seed):
            start_sec = time.perf_counter()
            result = self.function(self.settings)
            stop_sec = time.perf_counter()

        duration = timedelta(seconds=stop_sec - start_sec)

        metadata = ResultMetadata(
            start_datetime=start_datetime,
            duration=duration,
        )

        return Result(self.name, result, metadata)


# Register un-/structure hooks
converter.register_unstructure_hook(
    Benchmark,
    lambda o: dict(
        {"description": o.description},
        **make_dict_unstructure_fn(Benchmark, converter, function=override(omit=True))(
            o
        ),
    ),
)



"""Validation tests for constraints."""

import pytest
from pytest import param

from baybe.constraints.continuous import ContinuousCardinalityConstraint


@pytest.mark.parametrize(
    ("cardinalities", "error", "match"),
    [
        param(("0", 0), TypeError, "must be <class 'int'>", id="type_min"),
        param((0, "1"), TypeError, "must be <class 'int'>", id="type_max"),
        param((-1, 0), ValueError, "'min_cardinality' must be >= 0", id="loo_small"),
        param((1, 0), ValueError, "larger than the upper bound", id="wrong_order"),
        param((0, 3), ValueError, "exceed the number of parameters", id="too_large"),
        param((0, 2), ValueError, r"No constraint .* required", id="inactive"),
    ],
)
def test_invalid_cardinalities(cardinalities, error, match):
    """Providing an invalid parameter name raises an exception."""
    with pytest.raises(error, match=match):
        ContinuousCardinalityConstraint(["x", "y"], *cardinalities)



## Campaign Stopping

# Based on the insights from
# [this other example](/examples/Custom_Hooks/probability_of_improvement), we now
# demonstrate how to leverage the
# {func}`register_hooks <baybe.utils.basic.register_hooks>`
# mechanics to interrupt a running campaign based on a simple *Probability of
# Improvement (PI)* criterion. This approach could be used, for instance, to terminate
# unpromising campaigns early and refine their search spaces, or to end an ongoing
# optimization if the found results are sufficiently good.
# The underlying use case is taken from the example shown
# [here](/examples/Backtesting/full_lookup).

### Imports
import math
import os
import warnings

import pandas as pd
import seaborn as sns
import torch

from baybe import Campaign
from baybe.acquisition import ProbabilityOfImprovement
from baybe.exceptions import UnusedObjectWarning
from baybe.objectives import SingleTargetObjective
from baybe.objectives.base import Objective
from baybe.parameters import NumericalDiscreteParameter, SubstanceParameter
from baybe.recommenders import (
    BotorchRecommender,
    RandomRecommender,
    TwoPhaseMetaRecommender,
)
from baybe.searchspace import SearchSpace, SearchSpaceType
from baybe.simulation import simulate_scenarios
from baybe.targets import NumericalTarget
from baybe.utils import register_hooks
from baybe.utils.dataframe import to_tensor
from baybe.utils.plotting import create_example_plots

### Temporary
warnings.filterwarnings(
    "ignore", category=UnusedObjectWarning, message="explicit objective"
)
warnings.filterwarnings("ignore", category=DeprecationWarning)


### Settings

# Let's start by defining some basic settings required for the example:

SMOKE_TEST = "SMOKE_TEST" in os.environ
N_DOE_ITERATIONS = 2 if SMOKE_TEST else 25
N_MC_ITERATIONS = 2 if SMOKE_TEST else 20
N_INTERRUPTED_CAMPAIGNS = 2 if SMOKE_TEST else 5
BATCH_SIZE = 1
RANDOM_SEED = 1337


### Problem Definition and Lookup Functionality

# We load the dataframe containing the lookup data for the closed-loop simulation:

try:
    lookup = pd.read_excel("./../Backtesting/lookup.xlsx")
except FileNotFoundError:
    lookup = pd.read_excel("examples/Backtesting/lookup.xlsx")

# Following the setup described [here](../Backtesting/full_lookup.md), we create the
# building blocks for the optimization problem:

solvent_data = dict(set(zip(lookup.Solvent, lookup.Solvent_SMILES)))
base_data = dict(set(zip(lookup.Base, lookup.Base_SMILES)))
ligand_data = dict(set(zip(lookup.Ligand, lookup.Ligand_SMILES)))
temperature_values = set(lookup.Temp_C)
concentration_values = set(lookup.Concentration)

parameters = [
    SubstanceParameter(name="Solvent", data=solvent_data, encoding="MORDRED"),
    SubstanceParameter(name="Base", data=base_data, encoding="MORDRED"),
    SubstanceParameter(name="Ligand", data=ligand_data, encoding="MORDRED"),
    NumericalDiscreteParameter(name="Temp_C", values=temperature_values, tolerance=2),
    NumericalDiscreteParameter(name="Concentration", values=concentration_values),
]

searchspace = SearchSpace.from_product(parameters=parameters)

objective = SingleTargetObjective(target=NumericalTarget(name="yield", mode="MAX"))

recommender = TwoPhaseMetaRecommender(
    initial_recommender=RandomRecommender(), recommender=BotorchRecommender()
)

### Simulating the Uninterrupted Campaigns

# First, we run several Monte Carlo repetitions of the uninterrupted campaign to get a
# feeling for the average trajectory. For reproducibility, we also fix the random seed:

campaign = Campaign(searchspace, objective, recommender)
results_uninterrupted = simulate_scenarios(
    {"Average uninterrupted": campaign},
    lookup,
    batch_size=BATCH_SIZE,
    n_doe_iterations=N_DOE_ITERATIONS,
    n_mc_iterations=N_MC_ITERATIONS,
    random_seed=RANDOM_SEED,
)


### Defining the Campaign-Stopping Hook

# In order to interrupt a running campaign, we define a custom exception to identify our
# stopping event:


class CampaignStoppedException(Exception):
    """The campaign should be stopped."""


# Based on this exception class, we can now define a hook implementing the stopping
# criterion. For this purpose, we count the fraction of candidates with a PI exceeding a
# given value and terminate the campaign once the fraction falls below a certain
# threshold.

PI_THRESHOLD = 0.01  # PI of 1% to identify promising points
PI_REQUIRED_FRACTION = 0.2  # 20% of candidates must be above the threshold


def stop_on_PI(
    self: BotorchRecommender,
    searchspace: SearchSpace,
    objective: Objective | None = None,
    measurements: pd.DataFrame | None = None,
) -> None:
    """Raise an exception if the PI-based stopping criterion is fulfilled."""
    if searchspace.type != SearchSpaceType.DISCRETE:
        raise TypeError(
            f"Search spaces of type '{searchspace.type}' are not supported. "
            f"Currently, only search spaces of type '{SearchSpaceType.DISCRETE}' are "
            f"accepted."
        )
    acqf = ProbabilityOfImprovement()
    botorch_acqf = acqf.to_botorch(
        self._surrogate_model, searchspace, objective, measurements
    )
    _, candidates_comp_rep = searchspace.discrete.get_candidates()
    comp_rep_tensor = to_tensor(candidates_comp_rep).unsqueeze(1)
    acqf_values = botorch_acqf(comp_rep_tensor)

    n_pis_over = torch.sum(acqf_values > PI_THRESHOLD)
    n_pis_over_required = math.ceil(len(candidates_comp_rep) * PI_REQUIRED_FRACTION)
    if n_pis_over < n_pis_over_required:
        raise CampaignStoppedException(
            f"Less than {PI_REQUIRED_FRACTION*100:.0f}% of candidates are above the PI "
            f"threshold of {PI_THRESHOLD*100:.0f}% - Stopping the campaign."
        )


# Now, we attach the hook to the ``recommend`` function of our recommender class:

BotorchRecommender.recommend = register_hooks(
    BotorchRecommender.recommend, post_hooks=[stop_on_PI]
)

# ```{admonition} Monkeypatching
# :class: note
# The above monkeypatch registers the hook with all future instances of the recommender
# class. While it is possible to attach the hook only to a specific instance via
# ``MethodType`` (see [here](./basics.md)), this approach does not work well with the
# simulation utilities because they internally create deep copies of the simulated
# campaign, effectively bypassing the patch.
# ```

### Simulating the Interrupted Campaigns

# With the hook attached to the class, we again run several Monte Carlo repetitions of
# the same campaign. For this purpose, we instantiate a new recommender with the active
# hook and assign it to a fresh copy of the campaign:

recommender_with_hook = TwoPhaseMetaRecommender(
    initial_recommender=RandomRecommender(), recommender=BotorchRecommender()
)
campaign_with_hook = Campaign(searchspace, objective, recommender)

# Now, we can simply re-trigger the simulation loop. In order to establish a 1:1
# comparison, we use the same random seed as before so that the initial states of all
# trajectories are aligned with the previous runs:

results_interrupted = simulate_scenarios(
    {"Interrupted": campaign_with_hook},
    lookup,
    batch_size=BATCH_SIZE,
    n_doe_iterations=N_DOE_ITERATIONS,
    n_mc_iterations=N_INTERRUPTED_CAMPAIGNS,
    random_seed=RANDOM_SEED,
)

# ```{note}
# If an exception is thrown inside the loop, the function still returns the partial
# trajectory, which effectively implements early stopping.
# ```


### Plotting the Results

# Finally, we plot both the interrupted and the uninterrupted results.
# To display the latter in terms of individual trajectories, we can leverage the column
# that keeps track of the Monte Carlo iterations:

results_interrupted = results_interrupted.drop("Scenario", axis=1)
results_interrupted["Scenario"] = results_interrupted["Monte_Carlo_Run"].apply(
    lambda k: f"PI-stopped, run {k}"
)

# Now, we can easily create the plot from a single combined dataframe:

results = pd.concat([results_uninterrupted, results_interrupted])
ax = sns.lineplot(
    data=results,
    marker="o",
    markersize=10,
    x="Num_Experiments",
    y="yield_CumBest",
    hue="Scenario",
)
for line in ax.get_lines()[1:]:
    line.set_dashes([5, 2])
create_example_plots(ax=ax, base_name="campaign_stopping")



"""Tests for dataframe utilities."""

import numpy as np
import pandas as pd
import pytest

from baybe.utils.dataframe import add_noise_to_perturb_degenerate_rows


def test_degenerate_rows():
    """Test noise-based deduplication of degenerate rows."""
    # Create random dataframe
    df = pd.DataFrame(np.random.randint(0, 100, size=(5, 3))).astype(float)

    # Manually create some degenerate rows
    df.loc[1] = df.loc[0]  # Make row 1 identical to row 0
    df.loc[3] = df.loc[2]  # Make row 3 identical to row 2
    df.iloc[:, -1] = 50.0  # Make last column constant to test the edge case

    # Add noise
    add_noise_to_perturb_degenerate_rows(df)

    # Assert that the utility fixed the degenerate rows
    assert not df.duplicated().any(), "Degenerate rows were not fixed by the utility."


def test_degenerate_rows_invalid_input():
    """Test that the utility correctly handles invalid input."""
    # Create random dataframe
    df = pd.DataFrame(np.random.randint(0, 100, size=(5, 3))).astype(float)

    # Manually create some degenerate rows
    df.loc[1] = df.loc[0]  # Make row 1 identical to row 0
    df.loc[3] = df.loc[2]  # Make row 3 identical to row 2

    # Insert invalid data types
    df = df.astype(object)  # to avoid pandas column dtype warnings
    df["invalid"] = "A"
    df.iloc[1, 0] = "B"

    # Add noise
    with pytest.raises(TypeError):
        add_noise_to_perturb_degenerate_rows(df)



"""Temporary functionality for backward compatibility."""

from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

import cattrs
from cattrs.gen import make_dict_structure_fn

from baybe.serialization import converter
from baybe.targets.base import Target
from baybe.utils.basic import get_subclasses
from baybe.utils.boolean import is_abstract

if TYPE_CHECKING:
    from baybe.objectives.base import Objective


def structure_objective(val: dict, cls: type) -> Objective:
    """A structure hook that automatically determines an objective fallback type."""  # noqa: D401 (imperative mood)
    from baybe.objective import Objective as OldObjective
    from baybe.objectives.base import Objective
    from baybe.objectives.desirability import DesirabilityObjective
    from baybe.objectives.single import SingleTargetObjective

    val = val.copy()

    # Use the specified objective type
    try:
        if is_abstract(cls):
            _type = val.pop("type")
            try:
                cls = next(
                    cl for cl in get_subclasses(Objective) if cl.__name__ == _type
                )
            except StopIteration as ex:
                raise ValueError(f"Unknown subclass '{_type}'.") from ex
        fun = make_dict_structure_fn(
            cls,
            converter,
            _cattrs_forbid_extra_keys=True,
            _target=cattrs.override(rename="target"),
            _targets=cattrs.override(rename="targets"),
        )  # type: ignore
        return fun(val, cls)

    # If no type is provided, determine the type by the number of targets given
    except KeyError:
        n_targets = len(val["targets"])
        if n_targets == 0:
            raise ValueError("No targets are specified.")
        elif n_targets == 1:
            cls = SingleTargetObjective
        else:
            cls = DesirabilityObjective
        warnings.warn(
            f"An objective has been specified without a corresponding type. "
            f"Since {n_targets} target(s) have been provided, "
            f"'{cls.__name__}' is used as a fallback. "
            f"However, this behavior is deprecated and will be disabled in "
            f"a future version.",
            DeprecationWarning,
        )
        val["targets"] = converter.structure(val["targets"], list[Target])
        return OldObjective(**val)



"""Batch simulation of multiple campaigns."""

from __future__ import annotations

import warnings
from collections.abc import Callable
from copy import deepcopy
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Literal

import pandas as pd

from baybe.campaign import Campaign
from baybe.exceptions import NothingToSimulateError, UnusedObjectWarning
from baybe.simulation.core import simulate_experiment

if TYPE_CHECKING:
    from xarray import DataArray

_DEFAULT_SEED = 1337


def simulate_scenarios(
    scenarios: dict[Any, Campaign],
    lookup: pd.DataFrame | Callable[[pd.DataFrame], pd.DataFrame] | None = None,
    /,
    *,
    batch_size: int = 1,
    n_doe_iterations: int | None = None,
    initial_data: list[pd.DataFrame] | None = None,
    groupby: list[str] | None = None,
    n_mc_iterations: int = 1,
    random_seed: int | None = None,
    impute_mode: Literal[
        "error", "worst", "best", "mean", "random", "ignore"
    ] = "error",
    noise_percent: float | None = None,
) -> pd.DataFrame:
    """Simulate multiple Bayesian optimization scenarios.

    A wrapper function around :func:`baybe.simulation.core.simulate_experiment` that
    allows to specify multiple simulation settings at once.

    Args:
        scenarios: A dictionary mapping scenario identifiers to DOE specifications.
        lookup: See :func:`baybe.simulation.core.simulate_experiment`.
        batch_size: See :func:`baybe.simulation.core.simulate_experiment`.
        n_doe_iterations: See :func:`baybe.simulation.core.simulate_experiment`.
        initial_data: A list of initial data sets for which the scenarios should be
            simulated.
        groupby: The names of the parameters to be used to partition the search space.
            A separate simulation will be conducted for each partition, with the search
            restricted to that partition.
        n_mc_iterations: The number of Monte Carlo simulations to be used.
        random_seed: An optional integer specifying the random seed for the first Monte
            Carlo run. Each subsequent runs will increase this value by 1. If omitted,
            the current random seed is used.
        impute_mode: See :func:`baybe.simulation.core.simulate_experiment`.
        noise_percent: See :func:`baybe.simulation.core.simulate_experiment`.

    Returns:
        A dataframe like returned from :func:`baybe.simulation.core.simulate_experiment`
        but with additional columns. See the ``Note`` for details.

    Note:
        The following additional columns are contained in the dataframe returned by this
        function:

        * ``Scenario``: Specifies the scenario identifier of the respective simulation.
        * ``Monte_Carlo_Run``: Specifies the Monte Carlo repetition of the
          respective simulation.
        * Optional, if ``random_seed`` is provided: A column ``Random_Seed`` that
          specifies the random seed used for the respective simulation.
        * Optional, if ``initial_data`` is provided: A column ``Initial_Data`` that
          specifies the index of the initial data set used for the respective
          simulation.
        * Optional, if ``groupby`` is provided: A column for each ``groupby`` parameter
          that specifies the search space partition considered for the respective
          simulation.
    """

    @dataclass
    class SimulationResult:
        """A thin wrapper to enable dataframe-valued return values with xyzpy.

        Args:
            result: The result of the simulation.
        """

        result: pd.DataFrame

    def make_xyzpy_callable(result_variable: str) -> Callable:
        """Make a batch simulator that allows running campaigns in parallel."""
        from baybe._optional.simulation import xyzpy

        @xyzpy.label(var_names=[result_variable])
        def simulate(
            Scenario: str,
            Monte_Carlo_Run: int,
            Initial_Data=None,
        ):
            """Callable for xyzpy simulation."""
            data = None if initial_data is None else initial_data[Initial_Data]
            seed = None if random_seed is None else Monte_Carlo_Run + _DEFAULT_SEED
            result = _simulate_groupby(
                scenarios[Scenario],
                lookup,
                batch_size=batch_size,
                n_doe_iterations=n_doe_iterations,
                initial_data=data,
                groupby=groupby,
                random_seed=seed,
                impute_mode=impute_mode,
                noise_percent=noise_percent,
            )
            if random_seed is not None:
                result["Random_Seed"] = seed
            return SimulationResult(result)

        return simulate

    def unpack_simulation_results(array: DataArray) -> pd.DataFrame:
        """Turn the xyzpy simulation results into a flat dataframe."""
        # Convert to dataframe and remove the wrapper layer
        series = array.to_series()
        series = series.apply(lambda x: x.result)

        # Un-nest all simulation results
        dfs = []
        for setting, df_result in series.items():
            df_setting = pd.DataFrame(
                [setting], columns=series.index.names, index=df_result.index
            )
            dfs.append(pd.concat([df_setting, df_result], axis=1))

        # Concatenate all results into a single dataframe
        return pd.concat(dfs, ignore_index=True)

    # Collect the settings to be simulated
    combos = {"Scenario": scenarios.keys()}
    combos["Monte_Carlo_Run"] = range(n_mc_iterations)
    if initial_data:
        combos["Initial_Data"] = range(len(initial_data))

    # Simulate and unpack
    result_variable = "simulation_result"
    batch_simulator = make_xyzpy_callable(result_variable)

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            category=UnusedObjectWarning,
            module="baybe.recommenders.pure.nonpredictive.base",
        )
        da_results = batch_simulator.run_combos(combos)[result_variable]

    df_results = unpack_simulation_results(da_results)

    return df_results


def _simulate_groupby(
    campaign: Campaign,
    lookup: pd.DataFrame | Callable[[pd.DataFrame], pd.DataFrame] | None = None,
    /,
    *,
    batch_size: int = 1,
    n_doe_iterations: int | None = None,
    initial_data: pd.DataFrame | None = None,
    groupby: list[str] | None = None,
    random_seed: int = _DEFAULT_SEED,
    impute_mode: Literal[
        "error", "worst", "best", "mean", "random", "ignore"
    ] = "error",
    noise_percent: float | None = None,
) -> pd.DataFrame:
    """Scenario simulation for different search space partitions.

    A wrapper around :func:`baybe.simulation.core.simulate_experiment` that allows to
    partition the search space into different groups and run separate simulations for
    all groups where the search is restricted to the corresponding partition.

    Args:
        campaign: See :func:`baybe.simulation.core.simulate_experiment`.
        lookup: See :func:`baybe.simulation.core.simulate_experiment`.
        batch_size: See :func:`baybe.simulation.core.simulate_experiment`.
        n_doe_iterations: See :func:`baybe.simulation.core.simulate_experiment`.
        initial_data: See :func:`baybe.simulation.core.simulate_experiment`.
        groupby: See :func:`baybe.simulation.scenarios.simulate_scenarios`.
        random_seed: See :func:`baybe.simulation.core.simulate_experiment`.
        impute_mode: See :func:`baybe.simulation.core.simulate_experiment`.
        noise_percent: See :func:`baybe.simulation.core.simulate_experiment`.

    Returns:
        A dataframe like returned from
        :func:`baybe.simulation.core.simulate_experiment`, but with additional
        ``groupby columns`` (named according to the specified groupby parameters) that
        subdivide the results into the different simulations.

    Raises:
        NothingToSimulateError: If there is nothing to simulate.
    """
    # Create the groups. If no grouping is specified, use a single group containing
    # all parameter configurations.
    # NOTE: In the following, we intentionally work with *integer* indexing (iloc)
    #   instead of pandas indexes (loc), because the latter would yield wrong
    #   results in cases where the search space dataframe contains duplicate
    #   index entries (i.e., controlling the recommendable entries would affect
    #   all duplicates). While duplicate entries should be prevented by the search
    #   space constructor, the integer-based indexing provides a second safety net.
    #   Hence, the "reset_index" call.
    if groupby is None:
        groups = ((None, campaign.searchspace.discrete.exp_rep.reset_index()),)
    else:
        groups = campaign.searchspace.discrete.exp_rep.reset_index().groupby(groupby)

    # Simulate all subgroups
    dfs = []
    for group_id, group in groups:
        # Create a campaign that focuses only on the current group by excluding
        # off-group configurations from the candidates list
        # TODO: Reconsider if deepcopies are required once [16605] is resolved
        campaign_group = deepcopy(campaign)
        cols = [
            c
            for c in group.columns
            if c in campaign.searchspace.discrete.parameter_names
        ]
        campaign_group.toggle_discrete_candidates(
            group[cols], exclude=True, complement=True
        )

        # Run the group simulation
        try:
            df_group = simulate_experiment(
                campaign_group,
                lookup,
                batch_size=batch_size,
                n_doe_iterations=n_doe_iterations,
                initial_data=initial_data,
                random_seed=random_seed,
                impute_mode=impute_mode,
                noise_percent=noise_percent,
            )
        except NothingToSimulateError:
            continue

        # Add the group columns
        if groupby is not None:
            group_tuple = group_id if isinstance(group_id, tuple) else (group_id,)
            context = pd.DataFrame([group_tuple], columns=groupby, index=df_group.index)
            df_group = pd.concat([context, df_group], axis=1)

        dfs.append(df_group)

    # Collect all results
    if len(dfs) == 0:
        raise NothingToSimulateError
    df = pd.concat(dfs, ignore_index=True)

    return df



"""Continuous subspaces."""

from __future__ import annotations

import gc
import warnings
from collections.abc import Collection, Sequence
from itertools import chain
from typing import TYPE_CHECKING, Any, cast

import numpy as np
import pandas as pd
from attrs import define, field, fields
from typing_extensions import override

from baybe.constraints import (
    ContinuousCardinalityConstraint,
    ContinuousLinearConstraint,
)
from baybe.constraints.base import ContinuousConstraint, ContinuousNonlinearConstraint
from baybe.constraints.validation import (
    validate_cardinality_constraints_are_nonoverlapping,
)
from baybe.parameters import NumericalContinuousParameter
from baybe.parameters.base import ContinuousParameter
from baybe.parameters.utils import get_parameters_from_dataframe, sort_parameters
from baybe.searchspace.validation import (
    validate_parameter_names,
)
from baybe.serialization import SerialMixin, converter, select_constructor_hook
from baybe.utils.basic import to_tuple
from baybe.utils.dataframe import get_transform_objects, pretty_print_df
from baybe.utils.numerical import DTypeFloatNumpy
from baybe.utils.plotting import to_string

if TYPE_CHECKING:
    from baybe.searchspace.core import SearchSpace

_MAX_CARDINALITY_SAMPLING_ATTEMPTS = 10_000


@define
class SubspaceContinuous(SerialMixin):
    """Class for managing continuous subspaces.

    Builds the subspace from parameter definitions, keeps
    track of search metadata, and provides access to candidate sets and different
    parameter views.
    """

    parameters: tuple[NumericalContinuousParameter, ...] = field(
        converter=sort_parameters,
        validator=lambda _, __, x: validate_parameter_names(x),
    )
    """The parameters of the subspace."""

    constraints_lin_eq: tuple[ContinuousLinearConstraint, ...] = field(
        converter=to_tuple, factory=tuple
    )
    """Linear equality constraints."""

    constraints_lin_ineq: tuple[ContinuousLinearConstraint, ...] = field(
        converter=to_tuple, factory=tuple
    )
    """Linear inequality constraints."""

    constraints_nonlin: tuple[ContinuousNonlinearConstraint, ...] = field(
        converter=to_tuple, factory=tuple
    )
    """Nonlinear constraints."""

    @override
    def __str__(self) -> str:
        if self.is_empty:
            return ""

        # Convert the lists to dataFrames to be able to use pretty_printing
        param_list = [param.summary() for param in self.parameters]
        eq_constraints_list = [constr.summary() for constr in self.constraints_lin_eq]
        ineq_constraints_list = [
            constr.summary() for constr in self.constraints_lin_ineq
        ]
        nonlin_constraints_list = [
            constr.summary() for constr in self.constraints_nonlin
        ]
        param_df = pd.DataFrame(param_list)
        lin_eq_df = pd.DataFrame(eq_constraints_list)
        lin_ineq_df = pd.DataFrame(ineq_constraints_list)
        nonlinear_df = pd.DataFrame(nonlin_constraints_list)

        fields = [
            to_string(
                "Continuous Parameters", pretty_print_df(param_df, max_colwidth=None)
            ),
            to_string("Linear Equality Constraints", pretty_print_df(lin_eq_df)),
            to_string("Linear Inequality Constraints", pretty_print_df(lin_ineq_df)),
            to_string("Non-linear Constraints", pretty_print_df(nonlinear_df)),
        ]

        return to_string(self.__class__.__name__, *fields)

    @property
    def constraints_cardinality(self) -> tuple[ContinuousCardinalityConstraint, ...]:
        """Cardinality constraints."""
        return tuple(
            c
            for c in self.constraints_nonlin
            if isinstance(c, ContinuousCardinalityConstraint)
        )

    @constraints_lin_eq.validator
    def _validate_constraints_lin_eq(
        self, _, lst: list[ContinuousLinearConstraint]
    ) -> None:
        """Validate linear equality constraints."""
        # TODO Remove once eq and ineq constraints are consolidated into one list
        if not all(c.is_eq for c in lst):
            raise ValueError(
                f"The list '{fields(self.__class__).constraints_lin_eq.name}' of "
                f"{self.__class__.__name__} only accepts equality constraints, i.e. "
                f"the 'operator' for all list items should be '='."
            )

    @constraints_lin_ineq.validator
    def _validate_constraints_lin_ineq(
        self, _, lst: list[ContinuousLinearConstraint]
    ) -> None:
        """Validate linear inequality constraints."""
        # TODO Remove once eq and ineq constraints are consolidated into one list
        if any(c.is_eq for c in lst):
            raise ValueError(
                f"The list '{fields(self.__class__).constraints_lin_ineq.name}' of "
                f"{self.__class__.__name__} only accepts inequality constraints, i.e. "
                f"the 'operator' for all list items should be '>=' or '<='."
            )

    @constraints_nonlin.validator
    def _validate_constraints_nonlin(self, _, __) -> None:
        """Validate nonlinear constraints."""
        # Note: The passed constraints are accessed indirectly through the property
        validate_cardinality_constraints_are_nonoverlapping(
            self.constraints_cardinality
        )

    def to_searchspace(self) -> SearchSpace:
        """Turn the subspace into a search space with no discrete part."""
        from baybe.searchspace.core import SearchSpace

        return SearchSpace(continuous=self)

    @classmethod
    def empty(cls) -> SubspaceContinuous:
        """Create an empty continuous subspace."""
        return SubspaceContinuous(())

    @classmethod
    def from_parameter(cls, parameter: ContinuousParameter) -> SubspaceContinuous:
        """Create a subspace from a single parameter.

        Args:
            parameter: The parameter to span the subspace.

        Returns:
            The created subspace.
        """
        return cls.from_product([parameter])

    @classmethod
    def from_product(
        cls,
        parameters: Sequence[ContinuousParameter],
        constraints: Sequence[ContinuousConstraint] | None = None,
    ) -> SubspaceContinuous:
        """See :class:`baybe.searchspace.core.SearchSpace`."""
        constraints = constraints or []
        return SubspaceContinuous(
            parameters=[p for p in parameters if p.is_continuous],  # type:ignore[misc]
            constraints_lin_eq=[  # type:ignore[attr-misc]
                c
                for c in constraints
                if (isinstance(c, ContinuousLinearConstraint) and c.is_eq)
            ],
            constraints_lin_ineq=[  # type:ignore[attr-misc]
                c
                for c in constraints
                if (isinstance(c, ContinuousLinearConstraint) and not c.is_eq)
            ],
            constraints_nonlin=[  # type:ignore[attr-misc]
                c for c in constraints if isinstance(c, ContinuousNonlinearConstraint)
            ],
        )

    @classmethod
    def from_bounds(cls, bounds: pd.DataFrame) -> SubspaceContinuous:
        """Create a hyperrectangle-shaped continuous subspace with given bounds.

        Args:
            bounds: The bounds of the parameters.

        Returns:
            The constructed subspace.
        """
        # Assert that the input represents valid bounds
        assert bounds.shape[0] == 2
        assert (np.diff(bounds.values, axis=0) >= 0).all()
        assert bounds.apply(pd.api.types.is_numeric_dtype).all()

        # Create the corresponding parameters and from them the search space
        parameters = [
            NumericalContinuousParameter(cast(str, name), bound)
            for (name, bound) in bounds.items()
        ]
        return SubspaceContinuous(parameters)

    @classmethod
    def from_dataframe(
        cls,
        df: pd.DataFrame,
        parameters: Sequence[ContinuousParameter] | None = None,
    ) -> SubspaceContinuous:
        """Create a hyperrectangle-shaped continuous subspace from a dataframe.

        More precisely, create the smallest axis-aligned hyperrectangle-shaped
        continuous subspace that contains the points specified in the given dataframe.

        Args:
            df: The dataframe specifying the points spanning the subspace.
            parameters: Optional parameter objects corresponding to the columns in the
                given dataframe that can be provided to explicitly control parameter
                attributes. If a match between column name and parameter name is found,
                the corresponding parameter object is used. If a column has no match in
                the parameter list, a new
                :class:`baybe.parameters.numerical.NumericalContinuousParameter`
                is created with default optional arguments. For more details, see
                :func:`baybe.parameters.utils.get_parameters_from_dataframe`.

        Raises:
            ValueError: If parameter types other than
                :class:`baybe.parameters.numerical.NumericalContinuousParameter`
                are provided.

        Returns:
            The created continuous subspace.
        """
        # TODO: Add option for convex hull once constraints are in place

        if parameters and not all(
            isinstance(p, NumericalContinuousParameter) for p in parameters
        ):
            raise ValueError(
                "Currently, only parameters of type "
                "'{NumericalContinuousParameter.__name__}' are supported."
            )

        def continuous_parameter_factory(name: str, values: Collection[Any]):
            return NumericalContinuousParameter(name, (min(values), max(values)))

        # Get the full list of both explicitly and implicitly defined parameter
        parameters = get_parameters_from_dataframe(
            df, continuous_parameter_factory, parameters
        )

        return cls(parameters)

    @property
    def is_empty(self) -> bool:
        """Return whether this subspace is empty."""
        return len(self.parameters) == 0

    @property
    def parameter_names(self) -> tuple[str, ...]:
        """Return tuple of parameter names."""
        return tuple(p.name for p in self.parameters)

    @property
    def comp_rep_columns(self) -> tuple[str, ...]:
        """The columns spanning the computational representation."""
        return tuple(chain.from_iterable(p.comp_rep_columns for p in self.parameters))

    @property
    def comp_rep_bounds(self) -> pd.DataFrame:
        """The minimum and maximum values of the computational representation."""
        return pd.DataFrame(
            {p.name: p.bounds.to_tuple() for p in self.parameters},
            index=["min", "max"],
            dtype=DTypeFloatNumpy,
        )

    def _drop_parameters(self, parameter_names: Collection[str]) -> SubspaceContinuous:
        """Create a copy of the subspace with certain parameters removed.

        Args:
            parameter_names: The names of the parameter to be removed.

        Returns:
            The reduced subspace.
        """
        return SubspaceContinuous(
            parameters=[p for p in self.parameters if p.name not in parameter_names],
            constraints_lin_eq=[
                c._drop_parameters(parameter_names)
                for c in self.constraints_lin_eq
                if set(c.parameters) - set(parameter_names)
            ],
            constraints_lin_ineq=[
                c._drop_parameters(parameter_names)
                for c in self.constraints_lin_ineq
                if set(c.parameters) - set(parameter_names)
            ],
        )

    def transform(
        self,
        df: pd.DataFrame | None = None,
        /,
        *,
        allow_missing: bool = False,
        allow_extra: bool | None = None,
        data: pd.DataFrame | None = None,
    ) -> pd.DataFrame:
        """See :func:`baybe.searchspace.core.SearchSpace.transform`."""
        # >>>>>>>>>> Deprecation
        if not ((df is None) ^ (data is None)):
            raise ValueError(
                "Provide the data to be transformed as first positional argument."
            )

        if data is not None:
            df = data
            warnings.warn(
                "Providing the dataframe via the `data` argument is deprecated and "
                "will be removed in a future version. Please pass your dataframe "
                "as positional argument instead.",
                DeprecationWarning,
            )

        # Mypy does not infer from the above that `df` must be a dataframe here
        assert isinstance(df, pd.DataFrame)

        if allow_extra is None:
            allow_extra = True
            if set(df) - {p.name for p in self.parameters}:
                warnings.warn(
                    "For backward compatibility, the new `allow_extra` flag is set "
                    "to `True` when left unspecified. However, this behavior will be "
                    "changed in a future version. If you want to invoke the old "
                    "behavior, please explicitly set `allow_extra=True`.",
                    DeprecationWarning,
                )
        # <<<<<<<<<< Deprecation

        # Extract the parameters to be transformed
        parameters = get_transform_objects(
            df, self.parameters, allow_missing=allow_missing, allow_extra=allow_extra
        )

        # Transform the parameters
        return df[[p.name for p in parameters]]

    def samples_random(self, n_points: int = 1) -> pd.DataFrame:
        """Deprecated!"""  # noqa: D401
        warnings.warn(
            f"The method '{SubspaceContinuous.samples_random.__name__}' "
            f"has been deprecated and will be removed in a future version. "
            f"Use '{SubspaceContinuous.sample_uniform.__name__}' instead.",
            DeprecationWarning,
        )
        return self.sample_uniform(n_points)

    def sample_uniform(self, batch_size: int = 1) -> pd.DataFrame:
        """Draw uniform random parameter configurations from the continuous space.

        Args:
            batch_size: The number of parameter configurations to be sampled.

        Returns:
            A dataframe containing the parameter configurations as rows with columns
            corresponding to the parameter names.

        Raises:
            ValueError: If the subspace contains unsupported nonlinear constraints.
        """
        if not all(
            isinstance(c, ContinuousCardinalityConstraint)
            for c in self.constraints_nonlin
        ):
            raise ValueError(
                f"Currently, only nonlinear constraints of type "
                f"'{ContinuousCardinalityConstraint.__name__}' are supported."
            )

        if not self.parameters:
            return pd.DataFrame(index=pd.RangeIndex(0, batch_size))

        if (
            len(self.constraints_lin_eq) == 0
            and len(self.constraints_lin_ineq) == 0
            and len(self.constraints_cardinality) == 0
        ):
            return self._sample_from_bounds(batch_size, self.comp_rep_bounds.values)

        if len(self.constraints_cardinality) == 0:
            return self._sample_from_polytope(batch_size, self.comp_rep_bounds.values)

        return self._sample_from_polytope_with_cardinality_constraints(batch_size)

    def _sample_from_bounds(self, batch_size: int, bounds: np.ndarray) -> pd.DataFrame:
        """Draw uniform random samples over a hyperrectangle-shaped space."""
        points = np.random.uniform(
            low=bounds[0, :], high=bounds[1, :], size=(batch_size, len(self.parameters))
        )

        return pd.DataFrame(points, columns=self.parameter_names)

    def _sample_from_polytope(
        self, batch_size: int, bounds: np.ndarray
    ) -> pd.DataFrame:
        """Draw uniform random samples from a polytope."""
        import torch
        from botorch.utils.sampling import get_polytope_samples

        points = get_polytope_samples(
            n=batch_size,
            bounds=torch.from_numpy(bounds),
            equality_constraints=[
                c.to_botorch(self.parameters) for c in self.constraints_lin_eq
            ],
            inequality_constraints=[
                c.to_botorch(self.parameters) for c in self.constraints_lin_ineq
            ],
        )
        return pd.DataFrame(points, columns=self.parameter_names)

    def _sample_from_polytope_with_cardinality_constraints(
        self, batch_size: int
    ) -> pd.DataFrame:
        """Draw random samples from a polytope with cardinality constraints."""
        if not self.constraints_cardinality:
            raise RuntimeError(
                f"This method should not be called without any constraints of type "
                f"'{ContinuousCardinalityConstraint.__name__}' in place. "
                f"Use '{SubspaceContinuous._sample_from_bounds.__name__}' "
                f"or '{SubspaceContinuous._sample_from_polytope.__name__}' instead."
            )

        # List to store the created samples
        samples: list[pd.DataFrame] = []

        # Counter for failed sampling attempts
        n_fails = 0

        while len(samples) < batch_size:
            # Randomly set some parameters inactive
            inactive_params_sample = self._sample_inactive_parameters(1)[0]

            # Remove the inactive parameters from the search space
            subspace_without_cardinality_constraint = self._drop_parameters(
                inactive_params_sample
            )

            # Sample from the reduced space
            try:
                sample = subspace_without_cardinality_constraint.sample_uniform(1)
                samples.append(sample)
            except ValueError:
                n_fails += 1

            # Avoid infinite loop
            if n_fails >= _MAX_CARDINALITY_SAMPLING_ATTEMPTS:
                raise RuntimeError(
                    f"The number of failed sampling attempts has exceeded the limit "
                    f"of {_MAX_CARDINALITY_SAMPLING_ATTEMPTS}. "
                    f"It appears that the feasible region of the search space is very "
                    f"small. Please review the search space constraints."
                )

        # Combine the samples and fill in inactive parameters
        parameter_names = [p.name for p in self.parameters]
        return (
            pd.concat(samples, ignore_index=True)
            .reindex(columns=parameter_names)
            .fillna(0.0)
        )

    def _sample_inactive_parameters(self, batch_size: int = 1) -> list[set[str]]:
        """Sample inactive parameters according to the given cardinality constraints."""
        inactives_per_constraint = [
            con.sample_inactive_parameters(batch_size)
            for con in self.constraints_cardinality
        ]
        return [set(chain(*x)) for x in zip(*inactives_per_constraint)]

    def samples_full_factorial(self, n_points: int = 1) -> pd.DataFrame:
        """Deprecated!"""  # noqa: D401
        warnings.warn(
            f"The method '{SubspaceContinuous.samples_full_factorial.__name__}' "
            f"has been deprecated and will be removed in a future version. "
            f"Use '{SubspaceContinuous.sample_from_full_factorial.__name__}' instead.",
            DeprecationWarning,
        )
        return self.sample_from_full_factorial(n_points)

    def sample_from_full_factorial(self, batch_size: int = 1) -> pd.DataFrame:
        """Draw parameter configurations from the full factorial of the space.

        Args:
            batch_size: The number of parameter configurations to be sampled.

        Returns:
            A dataframe containing the parameter configurations as rows with columns
            corresponding to the parameter names.

        Raises:
            ValueError: If there are not enough points to sample from.
        """
        if len(full_factorial := self.full_factorial) < batch_size:
            raise ValueError(
                f"You are trying to sample {batch_size} points from the full factorial "
                f"of the continuous space bounds, but it has only "
                f"{len(full_factorial)} points."
            )

        return full_factorial.sample(n=batch_size).reset_index(drop=True)

    @property
    def full_factorial(self) -> pd.DataFrame:
        """Get the full factorial of the continuous space."""
        index = pd.MultiIndex.from_product(
            self.comp_rep_bounds.values.T.tolist(), names=self.parameter_names
        )

        return pd.DataFrame(index=index).reset_index()

    def get_parameters_by_name(
        self, names: Sequence[str]
    ) -> tuple[NumericalContinuousParameter, ...]:
        """Return parameters with the specified names.

        Args:
            names: Sequence of parameter names.

        Returns:
            The named parameters.
        """
        return tuple(p for p in self.parameters if p.name in names)


# Register deserialization hook
converter.register_structure_hook(SubspaceContinuous, select_constructor_hook)

# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



"""Optional ONNX import."""

from baybe.exceptions import OptionalImportError

try:
    import onnxruntime
except ModuleNotFoundError as ex:
    raise OptionalImportError(
        "Custom surrogate models are unavailable because 'onnxruntime' is not "
        "installed. "
        "Consider installing BayBE with 'onnx' dependency, "
        "e.g. via `pip install baybe[onnx]`."
    ) from ex

__all__ = [
    "onnxruntime",
]



"""Test Polars implementations of constraints."""

import pytest
from pandas.testing import assert_frame_equal

from baybe._optional.info import POLARS_INSTALLED
from baybe.searchspace.discrete import (
    _apply_constraint_filter_pandas,
    _apply_constraint_filter_polars,
    parameter_cartesian_prod_pandas,
    parameter_cartesian_prod_polars,
)

pytestmark = pytest.mark.skipif(
    not POLARS_INSTALLED, reason="Optional polars dependency not installed."
)


if POLARS_INSTALLED:
    import polars as pl


def _lazyframe_from_product(parameters):
    """Create a Polars lazyframe from the product of given parameters and return it."""
    param_frames = [pl.LazyFrame({p.name: p.values}) for p in parameters]

    # Handling edge cases
    if len(param_frames) == 1:
        return param_frames[0]

    # Cross-join parameters
    res = param_frames[0]
    for frame in param_frames[1:]:
        res = res.join(frame, how="cross", force_parallel=True)

    return res


@pytest.mark.parametrize("parameter_names", [["Fraction_1", "Fraction_2"]])
@pytest.mark.parametrize("constraint_names", [["Constraint_8"]])
def test_polars_prodsum1(parameters, constraints):
    """Tests Polars implementation of sum constraint."""
    ldf = _lazyframe_from_product(parameters)
    ldf, _ = _apply_constraint_filter_polars(ldf, constraints)

    # Number of entries with 1,2-sum above 150
    ldf = ldf.with_columns(sum=pl.sum_horizontal(["Fraction_1", "Fraction_2"]))
    ldf = ldf.filter(pl.col("sum") > 150)
    num_entries = len(ldf.collect())

    assert num_entries == 0


@pytest.mark.parametrize("parameter_names", [["Fraction_1", "Fraction_2"]])
@pytest.mark.parametrize("constraint_names", [["Constraint_9"]])
def test_polars_prodsum2(parameters, constraints):
    """Tests Polars implementation of product constrain."""
    ldf = _lazyframe_from_product(parameters)
    ldf, _ = _apply_constraint_filter_polars(ldf, constraints)

    # Number of entries with product under 30
    df = ldf.filter(
        pl.reduce(lambda acc, x: acc * x, pl.col(["Fraction_1", "Fraction_2"])).alias(
            "prod"
        )
        < 30
    ).collect()

    num_entries = len(df)
    assert num_entries == 0


@pytest.mark.parametrize("parameter_names", [["Fraction_1", "Fraction_2"]])
@pytest.mark.parametrize("constraint_names", [["Constraint_10"]])
def test_polars_prodsum3(parameters, constraints):
    """Tests Polars implementation of exact sum constraint."""
    ldf = _lazyframe_from_product(parameters)
    ldf, _ = _apply_constraint_filter_polars(ldf, constraints)

    # Number of entries with sum unequal to 100
    ldf = ldf.with_columns(sum=pl.sum_horizontal(["Fraction_1", "Fraction_2"]))
    df = ldf.select(abs(pl.col("sum") - 100)).filter(pl.col("sum") > 0.01).collect()

    num_entries = len(df)

    assert num_entries == 0


@pytest.mark.parametrize(
    "parameter_names",
    [["Solvent_1", "SomeSetting", "Temperature", "Pressure"]],
)
@pytest.mark.parametrize(
    "constraint_names", [["Constraint_4", "Constraint_5", "Constraint_6"]]
)
def test_polars_exclusion(mock_substances, parameters, constraints):
    """Tests Polars implementation of exclusion constraint."""
    ldf = _lazyframe_from_product(parameters)
    ldf, _ = _apply_constraint_filter_polars(ldf, constraints)

    # Number of entries with either first/second substance and a temperature above 151
    df = ldf.filter(
        (pl.col("Temperature") > 151)
        & (pl.col("Solvent_1").is_in(list(mock_substances)[:2]))
    ).collect()
    num_entries = len(df)
    assert num_entries == 0

    # Number of entries with either last / second last substance and a pressure above 5
    df = ldf.filter(
        (pl.col("Pressure") > 5)
        & (pl.col("Solvent_1").is_in(list(mock_substances)[-2:]))
    ).collect()
    num_entries = len(df)
    assert num_entries == 0

    # Number of entries with pressure below 3 and temperature above 120
    df = ldf.filter((pl.col("Pressure") < 3) & (pl.col("Temperature") > 120)).collect()
    num_entries = len(df)
    assert num_entries == 0


@pytest.mark.parametrize("parameter_names", [["Solvent_1", "Solvent_2", "Solvent_3"]])
@pytest.mark.parametrize("constraint_names", [["Constraint_7"]])
def test_polars_label_duplicates(parameters, constraints):
    """Tests Polars implementation of no-label duplicates constraint."""
    ldf = _lazyframe_from_product(parameters)
    ldf, _ = _apply_constraint_filter_polars(ldf, constraints)

    ldf = ldf.with_columns(
        pl.concat_list(pl.col(["Solvent_1", "Solvent_2", "Solvent_3"]))
        .list.eval(pl.element().n_unique())
        .explode()
        .alias("n_unique")
    )
    df = ldf.filter(pl.col("n_unique") != len(parameters)).collect()

    num_entries = len(df)
    assert num_entries == 0


@pytest.mark.parametrize("parameter_names", [["Solvent_1", "Solvent_2", "Solvent_3"]])
@pytest.mark.parametrize("constraint_names", [["Constraint_15"]])
def test_polars_linked_parameters(parameters, constraints):
    """Tests Polars implementation of linked parameters constraint."""
    ldf = _lazyframe_from_product(parameters)
    ldf, _ = _apply_constraint_filter_polars(ldf, constraints)

    ldf = ldf.with_columns(
        pl.concat_list(pl.col(["Solvent_1", "Solvent_2", "Solvent_3"]))
        .list.eval(pl.element().n_unique())
        .explode()
        .alias("n_unique")
    )
    df = ldf.filter(pl.col("n_unique") != 1).collect()

    num_entries = len(df)
    assert num_entries == 0


@pytest.mark.parametrize(
    "parameter_names",
    [
        [
            "Temperature",
            "Solvent_1",
            "Solvent_2",
            "Solvent_3",
            "Fraction_1",
            "Fraction_2",
            "Fraction_3",
        ]
    ],
)
@pytest.mark.parametrize(
    "constraint_names",
    [
        ["Constraint_4"],
        ["Constraint_12"],
        ["Constraint_15", "Constraint_8", "Constraint_9"],
    ],
)
def test_polars_product(constraints, parameters):
    """Test the result of parameter product and filtering."""
    # Do Polars product
    ldf = parameter_cartesian_prod_polars(parameters)
    df_pl = ldf.collect()

    # Do Pandas product
    df_pd = parameter_cartesian_prod_pandas(parameters)

    # Assert equality before filtering
    assert_frame_equal(df_pl.to_pandas(), df_pd)

    # Apply constraints
    df_pd_filtered = _apply_constraint_filter_pandas(df_pd, constraints)
    df_pl_filtered = (
        _apply_constraint_filter_polars(ldf, constraints)[0].collect().to_pandas()
    )

    # Assert order-agnostic equality of the two dataframes
    cols = df_pd_filtered.columns.tolist()
    assert_frame_equal(
        df_pd_filtered.sort_values(cols).reset_index(drop=True),
        df_pl_filtered.sort_values(cols).reset_index(drop=True),
    )



"""Executes the benchmarking module."""
# Run this via 'python -m benchmarks' from the root directory.

import os

from benchmarks.domains import BENCHMARKS
from benchmarks.persistence import (
    LocalFileObjectStorage,
    PathConstructor,
    S3ObjectStorage,
)

RUNS_IN_CI = "CI" in os.environ


def main():
    """Run all benchmarks."""
    for benchmark in BENCHMARKS:
        result = benchmark()
        path_constructor = PathConstructor.from_result(result)
        persist_dict = benchmark.to_dict() | result.to_dict()

        object_storage = S3ObjectStorage() if RUNS_IN_CI else LocalFileObjectStorage()
        object_storage.write_json(persist_dict, path_constructor)


if __name__ == "__main__":
    main()



"""Optional chemistry imports."""

from baybe.exceptions import OptionalImportError

try:
    from rdkit import Chem
    from skfp import fingerprints
    from skfp.bases import BaseFingerprintTransformer
    from skfp.preprocessing import ConformerGenerator, MolFromSmilesTransformer

except ModuleNotFoundError as ex:
    raise OptionalImportError(
        "Chemistry functionality is unavailable because the necessary optional "
        "dependencies are not installed. "
        "Consider installing BayBE with 'chem' dependency, "
        "e.g. via `pip install baybe[chem]`."
    ) from ex

__all__ = [
    "Chem",
    "fingerprints",
    "BaseFingerprintTransformer",
    "ConformerGenerator",
    "MolFromSmilesTransformer",
]



"""Continuous constraints."""

from __future__ import annotations

import gc
import math
from collections.abc import Collection, Sequence
from typing import TYPE_CHECKING, Any

import numpy as np
from attr.validators import in_
from attrs import define, field

from baybe.constraints.base import (
    CardinalityConstraint,
    ContinuousConstraint,
    ContinuousNonlinearConstraint,
)
from baybe.parameters import NumericalContinuousParameter
from baybe.utils.numerical import DTypeFloatNumpy
from baybe.utils.validation import finite_float

if TYPE_CHECKING:
    from torch import Tensor

_valid_linear_constraint_operators = ["=", ">=", "<="]


@define
class ContinuousLinearConstraint(ContinuousConstraint):
    """Class for continuous linear constraints.

    Continuous linear constraints use parameter lists and coefficients to define
    in-/equality constraints over a continuous parameter space.
    """

    # object variables
    operator: str = field(validator=in_(_valid_linear_constraint_operators))
    """Defines the operator used in the equation. Internally this will negate rhs and
    coefficients for `<=`."""

    coefficients: list[float] = field()
    """In-/equality coefficient for each entry in ``parameters``."""

    rhs: float = field(default=0.0, converter=float, validator=finite_float)
    """Right-hand side value of the in-/equality."""

    @coefficients.validator
    def _validate_coefficients(  # noqa: DOC101, DOC103
        self, _: Any, coefficients: list[float]
    ) -> None:
        """Validate the coefficients.

        Raises:
            ValueError: If the number of coefficients does not match the number of
                parameters.
        """
        if len(self.parameters) != len(coefficients):
            raise ValueError(
                "The given 'coefficients' list must have one floating point entry for "
                "each entry in 'parameters'."
            )

    @coefficients.default
    def _default_coefficients(self):
        """Return equal weight coefficients as default."""
        return [1.0] * len(self.parameters)

    @property
    def _multiplier(self) -> float:
        """The internal multiplier for rhs and coefficients."""
        return -1.0 if self.operator == "<=" else 1.0

    @property
    def is_eq(self):
        """Whether this constraint models an equality (assumed inequality otherwise)."""
        return self.operator == "="

    def _drop_parameters(
        self, parameter_names: Collection[str]
    ) -> ContinuousLinearConstraint:
        """Create a copy of the constraint with certain parameters removed.

        Args:
            parameter_names: The names of the parameter to be removed.

        Returns:
            The reduced constraint.
        """
        parameters = [p for p in self.parameters if p not in parameter_names]
        coefficients = [
            c
            for p, c in zip(self.parameters, self.coefficients)
            if p not in parameter_names
        ]
        return ContinuousLinearConstraint(
            parameters, self.operator, coefficients, self.rhs
        )

    def to_botorch(
        self, parameters: Sequence[NumericalContinuousParameter], idx_offset: int = 0
    ) -> tuple[Tensor, Tensor, float]:
        """Cast the constraint in a format required by botorch.

        Used in calling ``optimize_acqf_*`` functions, for details see
        https://botorch.org/api/optim.html#botorch.optim.optimize.optimize_acqf

        Args:
            parameters: The parameter objects of the continuous space.
            idx_offset: Offset to the provided parameter indices.

        Returns:
            The tuple required by botorch.
        """
        import torch

        from baybe.utils.torch import DTypeFloatTorch

        param_names = [p.name for p in parameters]
        param_indices = [
            param_names.index(p) + idx_offset
            for p in self.parameters
            if p in param_names
        ]

        return (
            torch.tensor(param_indices),
            torch.tensor(
                [self._multiplier * c for c in self.coefficients], dtype=DTypeFloatTorch
            ),
            np.asarray(self._multiplier * self.rhs, dtype=DTypeFloatNumpy).item(),
        )


@define
class ContinuousCardinalityConstraint(
    CardinalityConstraint, ContinuousNonlinearConstraint
):
    """Class for continuous cardinality constraints."""

    def sample_inactive_parameters(self, batch_size: int = 1) -> list[set[str]]:
        """Sample sets of inactive parameters according to the cardinality constraints.

        Args:
            batch_size: The number of parameter sets to be sampled.

        Returns:
            A list of sampled inactive parameter sets, where each set holds the
            corresponding parameter names.
        """
        # The number of possible parameter configuration per set cardinality
        n_configurations_per_cardinality = [
            math.comb(len(self.parameters), n)
            for n in range(self.min_cardinality, self.max_cardinality + 1)
        ]

        # Probability of each set cardinality under the assumption that all possible
        # inactive parameter sets are equally likely
        probabilities = n_configurations_per_cardinality / np.sum(
            n_configurations_per_cardinality
        )

        # Sample the number of active/inactive parameters
        n_active_params = np.random.choice(
            np.arange(self.min_cardinality, self.max_cardinality + 1),
            batch_size,
            p=probabilities,
        )
        n_inactive_params = len(self.parameters) - n_active_params

        # Sample the inactive parameters
        inactive_params = [
            set(np.random.choice(self.parameters, n_inactive, replace=False))
            for n_inactive in n_inactive_params
        ]

        return inactive_params


# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



"""Custom exceptions and warnings."""


##### Warnings #####


class UnusedObjectWarning(UserWarning):
    """
    A method or function was called with undesired arguments which indicates an
    unintended user fault.
    """


##### Exceptions #####


class IncompatibilityError(Exception):
    """Incompatible components are used together."""


class IncompatibleSearchSpaceError(IncompatibilityError):
    """
    A recommender is used with a search space that contains incompatible parts,
    e.g. a discrete recommender is used with a hybrid or continuous search space.
    """


class IncompatibleAcquisitionFunctionError(IncompatibilityError):
    """An incompatible acquisition function was selected."""


class IncompatibleExplainerError(IncompatibilityError):
    """An explainer is incompatible with the data it is presented."""


class IncompatibleArgumentError(IncompatibilityError):
    """An incompatible argument was passed to a callable."""


class NotEnoughPointsLeftError(Exception):
    """
    More recommendations are requested than there are viable parameter configurations
    left in the search space.
    """


class NoMCAcquisitionFunctionError(Exception):
    """
    A Monte Carlo acquisition function is required but an analytical acquisition
    function has been selected by the user.
    """


class EmptySearchSpaceError(Exception):
    """The created search space contains no parameters."""


class NothingToSimulateError(Exception):
    """There is nothing to simulate because there are no testable configurations."""


class NoRecommendersLeftError(Exception):
    """A recommender is requested by a meta recommender but there are no recommenders
    left.
    """


class NumericalUnderflowError(Exception):
    """A computation would lead to numerical underflow."""


class OptionalImportError(ImportError):
    """An attempt was made to import an optional but uninstalled dependency."""


class DeprecationError(Exception):
    """Signals the use of a deprecated mechanism to the user, interrupting execution."""


class UnidentifiedSubclassError(Exception):
    """A specified subclass cannot be found in the given class hierarchy."""


class ModelNotTrainedError(Exception):
    """A prediction/transformation is attempted before the model has been trained."""


class UnmatchedAttributeError(Exception):
    """An attribute cannot be matched against a certain callable signature."""


class InvalidSurrogateModelError(Exception):
    """An invalid surrogate model was chosen."""


class InvalidTargetValueError(Exception):
    """A target value was entered that is not in the target space."""



"""NGBoost surrogates."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar

from attrs import define, field
from ngboost import NGBRegressor
from typing_extensions import override

from baybe.parameters.base import Parameter
from baybe.surrogates.base import IndependentGaussianSurrogate
from baybe.surrogates.utils import batchify_mean_var_prediction, catch_constant_targets
from baybe.surrogates.validation import get_model_params_validator
from baybe.utils.plotting import to_string

if TYPE_CHECKING:
    from botorch.models.transforms.input import InputTransform
    from botorch.models.transforms.outcome import OutcomeTransform
    from torch import Tensor


@catch_constant_targets
@define
class NGBoostSurrogate(IndependentGaussianSurrogate):
    """A natural-gradient-boosting surrogate model."""

    supports_transfer_learning: ClassVar[bool] = False
    # See base class.

    _default_model_params: ClassVar[dict] = {"n_estimators": 25, "verbose": False}
    """Class variable encoding the default model parameters."""

    model_params: dict[str, Any] = field(
        factory=dict,
        converter=dict,
        validator=get_model_params_validator(NGBRegressor.__init__),
    )
    """Optional model parameter that will be passed to the surrogate constructor."""

    _model: NGBRegressor | None = field(init=False, default=None, eq=False)
    """The actual model."""

    def __attrs_post_init__(self):
        self.model_params = {**self._default_model_params, **self.model_params}

    @override
    @staticmethod
    def _make_parameter_scaler_factory(
        parameter: Parameter,
    ) -> type[InputTransform] | None:
        # Tree-like models do not require any input scaling
        return None

    @override
    @staticmethod
    def _make_target_scaler_factory() -> type[OutcomeTransform] | None:
        # Tree-like models do not require any output scaling
        return None

    @override
    @batchify_mean_var_prediction
    def _estimate_moments(
        self, candidates_comp_scaled: Tensor, /
    ) -> tuple[Tensor, Tensor]:
        # FIXME[typing]: It seems there is currently no better way to inform the type
        #   checker that the attribute is available at the time of the function call
        assert self._model is not None

        import torch

        # Get predictions
        dists = self._model.pred_dist(candidates_comp_scaled)

        # Split into posterior mean and variance
        mean = torch.from_numpy(dists.mean())
        var = torch.from_numpy(dists.var)

        return mean, var

    @override
    def _fit(self, train_x: Tensor, train_y: Tensor) -> None:
        self._model = NGBRegressor(**(self.model_params)).fit(train_x, train_y.ravel())

    @override
    def __str__(self) -> str:
        fields = [to_string("Model Params", self.model_params, single_line=True)]
        return to_string(super().__str__(), *fields)



"""Validation utilities."""

from __future__ import annotations

import math
from collections.abc import Callable
from typing import Any

from attrs import Attribute


def validate_not_nan(self: Any, attribute: Attribute, value: Any) -> None:
    """Attrs-compatible validator to forbid 'nan' values."""
    if isinstance(value, float) and math.isnan(value):
        raise ValueError(
            f"The value passed to attribute '{attribute.name}' of class "
            f"'{self.__class__.__name__}' cannot be 'nan'."
        )


def _make_restricted_float_validator(
    allow_nan: bool, allow_inf: bool
) -> Callable[[Any, Attribute, Any], None]:
    """Make an attrs-compatible validator for restricted floats.

    Args:
        allow_nan: If False, validated values cannot be 'nan'.
        allow_inf: If False, validated values cannot be 'inf' or '-inf'.

    Raises:
        ValueError: If no float range restriction is in place.

    Returns:
        The validator.
    """
    if allow_nan and allow_inf:
        raise ValueError(
            "The requested validator would not restrict the float range. "
            "Hence, you can use `attrs.validators.instance_of(float)` instead."
        )

    def validator(self: Any, attribute: Attribute, value: Any) -> None:
        if not isinstance(value, float):
            raise ValueError(
                f"Values assigned to attribute '{attribute.name}' of class "
                f"'{self.__class__.__name__}' must be of type 'float'. "
                f"Given: {value} (type: {type(value)})"
            )
        if not allow_inf and math.isinf(value):
            raise ValueError(
                f"Values assigned to attribute '{attribute.name}' of class "
                f"'{self.__class__.__name__}' cannot be 'inf' or '-inf'."
            )
        if not allow_nan and math.isnan(value):
            raise ValueError(
                f"Values assigned to attribute '{attribute.name}' of class "
                f"'{self.__class__.__name__}' cannot be 'nan'."
            )

    return validator


finite_float = _make_restricted_float_validator(allow_nan=False, allow_inf=False)
"""Validator for finite (i.e., non-nan and non-infinite) floats."""

non_nan_float = _make_restricted_float_validator(allow_nan=False, allow_inf=True)
"""Validator for non-nan floats."""

non_inf_float = _make_restricted_float_validator(allow_nan=True, allow_inf=False)
"""Validator for non-infinite floats."""



"""Chemistry tools."""

import os
import ssl
import tempfile
import urllib.request
import warnings
from collections.abc import Sequence
from functools import lru_cache
from pathlib import Path

import numpy as np
import pandas as pd
from joblib import Memory

from baybe._optional.chem import (
    BaseFingerprintTransformer,
    Chem,
    ConformerGenerator,
    MolFromSmilesTransformer,
    fingerprints,
)
from baybe.parameters.enum import SubstanceEncoding
from baybe.utils.numerical import DTypeFloatNumpy

# Caching
_cachedir = os.environ.get(
    "BAYBE_CACHE_DIR", str(Path(tempfile.gettempdir()) / ".baybe_cache")
)


def _dummy_wrapper(func):
    return func


_disk_cache = (
    _dummy_wrapper if _cachedir == "" else Memory(Path(_cachedir), verbose=0).cache
)


def name_to_smiles(name: str) -> str:
    """Convert from chemical name to SMILES string using chemical identifier resolver.

    This script is useful to combine with ``df.apply`` from pandas, hence it does not
    throw exceptions for invalid molecules but instead returns an empty string for
    easy subsequent postprocessing of the dataframe.

    Args:
        name: Name or nickname of compound.

    Returns:
        SMILES string corresponding to chemical name.
    """
    name = name.replace(" ", "%20")

    try:
        url = "http://cactus.nci.nih.gov/chemical/structure/" + name + "/smiles"
        ctx = ssl.create_default_context()
        ctx.check_hostname = False
        ctx.verify_mode = ssl.CERT_NONE

        with urllib.request.urlopen(url, context=ctx) as web:
            smiles = web.read().decode("utf8")

        smiles = str(smiles)
        if "</div>" in smiles:
            return ""

        return smiles

    except Exception:
        return ""


@lru_cache(maxsize=None)
@_disk_cache
def _molecule_to_fingerprint_features(
    molecule: str | Chem.Mol,
    encoder: BaseFingerprintTransformer,
) -> np.ndarray:
    """Compute molecular fingerprint for a single molecule.

    Args:
        molecule: SMILES string or molecule object.
        encoder: Instance of the fingerprint class to be used for computation.

    Returns:
        Array of fingerprint features.
    """
    return encoder.transform([molecule])


def smiles_to_fingerprint_features(
    smiles: Sequence[str],
    encoding: SubstanceEncoding,
    prefix: str | None = None,
    kwargs_conformer: dict | None = None,
    kwargs_fingerprint: dict | None = None,
) -> pd.DataFrame:
    """Compute molecular fingerprints for a list of SMILES strings.

    Args:
        smiles: Sequence of SMILES strings.
        encoding: Encoding used to transform SMILES to fingerprints.
        prefix: Name prefix for each descriptor (e.g., nBase --> <prefix>_nBase).
        kwargs_conformer: kwargs for conformer generator
        kwargs_fingerprint: kwargs for fingerprint generator

    Returns:
        Dataframe containing fingerprints for each SMILES string.
    """
    kwargs_fingerprint = kwargs_fingerprint or {}
    kwargs_conformer = kwargs_conformer or {}

    if encoding is SubstanceEncoding.MORGAN_FP:
        warnings.warn(
            f"Substance encoding '{encoding.name}' is deprecated and will be disabled "
            f"in a future version. Use '{SubstanceEncoding.ECFP.name}' "
            f"with 'fp_size' 1024 and 'radius' 4 instead.",
            DeprecationWarning,
        )
        encoding = SubstanceEncoding.ECFP
        kwargs_fingerprint.update({"fp_size": 1024, "radius": 4})

    elif encoding is SubstanceEncoding.RDKIT:
        warnings.warn(
            f"Substance encoding '{encoding.name}' is deprecated and will be disabled "
            f"in a future version. Use '{SubstanceEncoding.RDKIT2DDESCRIPTORS.name}' "
            f"instead.",
            DeprecationWarning,
        )
        encoding = SubstanceEncoding.RDKIT2DDESCRIPTORS

    fingerprint_cls = get_fingerprint_class(encoding)
    fingerprint_encoder = fingerprint_cls(**kwargs_fingerprint)

    if fingerprint_encoder.requires_conformers:
        mol_list = ConformerGenerator(**kwargs_conformer).transform(
            MolFromSmilesTransformer().transform(smiles)
        )
    else:
        mol_list = smiles

    features = np.concatenate(
        [
            _molecule_to_fingerprint_features(mol, fingerprint_encoder)
            for mol in mol_list
        ]
    )
    name = f"{encoding.name}_"
    prefix = prefix + "_" if prefix else ""
    feature_names_out = fingerprint_encoder.get_feature_names_out()
    no_descriptor_names = all("fingerprint" in f for f in feature_names_out)
    suffixes = [
        f.split("fingerprint")[1] if no_descriptor_names else f
        for f in feature_names_out
    ]
    col_names = [prefix + name + suffix for suffix in suffixes]
    df = pd.DataFrame(features, columns=col_names, dtype=DTypeFloatNumpy)

    return df


def get_fingerprint_class(encoding: SubstanceEncoding) -> BaseFingerprintTransformer:
    """Retrieve the fingerprint class corresponding to a given encoding.

    Args:
        encoding: A substance encoding.

    Raises:
        ValueError: If no fingerprint class for the specified encoding is found.

    Returns:
        The fingerprint class.
    """
    # Exception case
    if encoding is SubstanceEncoding.RDKITFINGERPRINT:
        return fingerprints.RDKitFingerprint

    try:
        cls_name = next(
            name
            for name in dir(fingerprints)
            if (encoding.name + "Fingerprint").casefold() == name.casefold()
        )
    except StopIteration as e:
        raise ValueError(
            f"No fingerprint class exists for the specified encoding '{encoding.name}'."
        ) from e
    return getattr(fingerprints, cls_name)


def get_canonical_smiles(smiles: str) -> str:
    """Return the "canonical" representation of the given SMILES."""
    try:
        return Chem.MolToSmiles(Chem.MolFromSmiles(smiles))
    except Exception:
        raise ValueError(f"The SMILES '{smiles}' does not appear to be valid.")



"""Validation tests for targets."""

import pytest
from pytest import param

from baybe.targets.binary import BinaryTarget
from baybe.targets.numerical import NumericalTarget


@pytest.mark.parametrize(
    ("mode", "bounds"),
    [
        param("MATCH", None, id="non_closed_match_mode"),
        param("MAX", (0, None), id="half_open"),
        param("MAX", (0, 0), id="degenerate"),
    ],
)
def test_numerical_target_invalid_bounds_mode(mode, bounds):
    """Providing invalid bounds raises an exception."""
    with pytest.raises(ValueError):
        NumericalTarget(name="invalid_bounds", mode=mode, bounds=bounds)


@pytest.mark.parametrize(
    ("mode", "bounds", "transformation"),
    [
        param("MIN", None, "BELL", id="bell_for_min"),
        param("MATCH", (0, 1), "LINEAR", id="linear_for_match"),
    ],
)
def test_numerical_target_incompatible_transform_mode(mode, bounds, transformation):
    with pytest.raises(ValueError):
        NumericalTarget(
            name="incompatible_transform",
            mode=mode,
            bounds=bounds,
            transformation=transformation,
        )


@pytest.mark.parametrize(
    ("choices", "error", "match"),
    [
        param((None, 0), TypeError, "'success_value' must be", id="wrong_type"),
        param((0, 0), ValueError, "must be different", id="identical"),
    ],
)
def test_binary_target_invalid_values(choices, error, match):
    """Providing invalid choice values raises an error."""
    with pytest.raises(error, match=match):
        BinaryTarget(
            name="invalid_value",
            success_value=choices[0],
            failure_value=choices[1],
        )


@pytest.mark.parametrize(
    "mode",
    ["MIN", "MAX"],
)
def test_providing_transformation_without_bounds(mode):
    """Providing a transformation without bounds raises an error."""
    with pytest.raises(ValueError, match="but did not specify any bounds."):
        NumericalTarget(
            name="transformation_without_bounds", mode=mode, transformation="LINEAR"
        )



"""Functionality for transfer learning backtesting."""

from __future__ import annotations

from copy import deepcopy
from typing import Any

import pandas as pd

from baybe.campaign import Campaign
from baybe.parameters import TaskParameter
from baybe.searchspace import SearchSpaceType
from baybe.simulation.scenarios import simulate_scenarios


def simulate_transfer_learning(
    campaign: Campaign,
    lookup: pd.DataFrame,
    /,
    *,
    batch_size: int = 1,
    n_doe_iterations: int | None = None,
    groupby: list[str] | None = None,
    n_mc_iterations: int = 1,
) -> pd.DataFrame:
    """Simulate Bayesian optimization with transfer learning.

    A wrapper around :func:`baybe.simulation.scenarios.simulate_scenarios` that
    partitions the search space into its tasks and simulates each task with the training
    data from the remaining tasks.

    **NOTE:**
    Currently, the simulation only supports purely discrete search spaces. This is
    because ``lookup`` serves both as the loop-closing element **and** as the source
    for off-task training data. For continuous (or mixed) spaces, the lookup mechanism
    would need to be either implemented as a callable (in which case the training data
    must be provided separately) or the continuous parameters need to be effectively
    restricted to the finite number of provided lookup configurations. Neither is
    implemented at the moment.

    Args:
        campaign: See :func:`baybe.simulation.core.simulate_experiment`.
        lookup: See :func:`baybe.simulation.scenarios.simulate_scenarios`.
        batch_size: See :func:`baybe.simulation.scenarios.simulate_scenarios`.
        n_doe_iterations: See :func:`baybe.simulation.scenarios.simulate_scenarios`.
        groupby: See :func:`baybe.simulation.scenarios.simulate_scenarios`.
        n_mc_iterations: See :func:`baybe.simulation.scenarios.simulate_scenarios`.

    Returns:
        A dataframe as returned by :func:`baybe.simulation.scenarios.simulate_scenarios`
        where the different tasks are represented in the ``Scenario`` column.

    Raises:
        NotImplementedError: If a non-discrete search space is chosen.
    """
    # TODO: Currently, we assume a purely discrete search space
    if campaign.searchspace.type != SearchSpaceType.DISCRETE:
        raise NotImplementedError(
            "Currently, only purely discrete search spaces are supported. "
            "For details, see NOTE in the function docstring."
        )

    # TODO [16932]: Currently, we assume exactly one task parameter exists
    # Extract the single task parameter
    task_params = [p for p in campaign.parameters if isinstance(p, TaskParameter)]
    if len(task_params) > 1:
        raise NotImplementedError(
            "Currently, transfer learning supports only a single task parameter."
        )
    task_param = task_params[0]

    # Create simulation objects for all tasks
    scenarios: dict[Any, Campaign] = {}
    for task in task_param.values:
        # Create a campaign that focuses only on the current task by excluding
        # off-task configurations from the candidates list
        # TODO: Reconsider if deepcopies are required once [16605] is resolved
        campaign_task = deepcopy(campaign)
        campaign_task.toggle_discrete_candidates(
            pd.DataFrame({task_param.name: [task]}), exclude=True, complement=True
        )

        # Use all off-task data as training data
        df_train = lookup[lookup[task_param.name] != task]
        campaign_task.add_measurements(df_train)

        # Add the task scenario
        scenarios[task] = campaign_task

    # Simulate all tasks
    return simulate_scenarios(
        scenarios,
        lookup,
        batch_size=batch_size,
        n_doe_iterations=n_doe_iterations,
        groupby=groupby,
        n_mc_iterations=n_mc_iterations,
        impute_mode="ignore",
    )



"""Base functionality for all BayBE targets."""

from __future__ import annotations

import gc
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

import pandas as pd
from attrs import define, field
from typing_extensions import override

from baybe.serialization import (
    SerialMixin,
    converter,
    get_base_structure_hook,
    unstructure_base,
)

if TYPE_CHECKING:
    from baybe.objective import SingleTargetObjective


@define(frozen=True)
class Target(ABC, SerialMixin):
    """Abstract base class for all target variables.

    Stores information about the range, transformations, etc.
    """

    name: str = field()
    """The name of the target."""

    def to_objective(self) -> SingleTargetObjective:
        """Create a single-task objective from the target."""
        from baybe.objectives.single import SingleTargetObjective

        return SingleTargetObjective(self)

    @abstractmethod
    def transform(self, series: pd.Series, /) -> pd.Series:
        """Transform target measurements to computational representation.

        Args:
            series: The target measurements in experimental representation to be
                transformed.

        Returns:
            A series containing the transformed measurements. The series name matches
            that of the input.
        """

    @abstractmethod
    def summary(self) -> dict:
        """Return a custom summarization of the target."""

    @override
    def __str__(self) -> str:
        return str(self.summary())


# Register (un-)structure hooks
converter.register_structure_hook(Target, get_base_structure_hook(Target))
converter.register_unstructure_hook(Target, unstructure_base)

# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



## Example for using the multi target mode for the objective

# Example for using the multi target mode for the objective.
# It uses a desirability value to handle several targets.

# This example assumes some basic familiarity with using BayBE.
# We thus refer to [`campaign`](./../Basics/campaign.md) for a basic example.

### Necessary imports for this example

import pandas as pd

from baybe import Campaign
from baybe.objectives import DesirabilityObjective
from baybe.parameters import CategoricalParameter, NumericalDiscreteParameter
from baybe.searchspace import SearchSpace
from baybe.targets import NumericalTarget
from baybe.utils.dataframe import add_fake_measurements

### Experiment setup and creating the searchspace

Categorical_1 = CategoricalParameter("Cat_1", values=["22", "33"], encoding="OHE")
Categorical_2 = CategoricalParameter(
    "Cat_2",
    values=["very bad", "bad", "OK", "good", "very good"],
    encoding="INT",
)
Num_disc_1 = NumericalDiscreteParameter(
    "Num_disc_1", values=[1, 2, 3, 4, 6, 8, 10], tolerance=0.3
)
Num_disc_2 = NumericalDiscreteParameter(
    "Num_disc_2", values=[-1, -3, -6, -9], tolerance=0.3
)

parameters = [Categorical_1, Categorical_2, Num_disc_1, Num_disc_2]

searchspace = SearchSpace.from_product(parameters=parameters)


### Defining the targets

# The multi target mode is handled when creating the objective object.
# Thus we first need to define the different targets.

# This examples has different targets with different modes.
# The first target is maximized and while the second one is minimized.
# Note that in this multi target mode, the user must specify bounds for each target.

Target_1 = NumericalTarget(
    name="Target_1", mode="MAX", bounds=(0, 100), transformation="LINEAR"
)
Target_2 = NumericalTarget(
    name="Target_2", mode="MIN", bounds=(0, 100), transformation="LINEAR"
)

# For each target it is also possible to specify a `target_transform` function.
# A detailed discussion of this functionality can be found at the end of this example.

# In this example, define a third target working with the mode `MATCH`.
# We furthermore use `target_transform="BELL"`.

Target_3 = NumericalTarget(
    name="Target_3", mode="MATCH", bounds=(45, 55), transformation="BELL"
)

# Note that the `MATCH` mode seeks to have the target at the mean between the two bounds.
# For example, choosing 95 and 105 will lead the algorithm seeking 100 as the optimal value.
# Thus, using the bounds, it is possible to control both the match target and
# the range around this target that is considered viable.


### Creating the objective

# Now to work with these three targets the objective object must be properly created.
# The mode is set to `DESIRABILITY` and the targets are described in a list.

targets = [Target_1, Target_2, Target_3]

# As the recommender requires a single function, the different targets need to be combined.
# Thus, a `scalarizer` is used to create a single target out of the several targets given.
# The combine function can either be the mean `MEAN` or the geometric mean `GEOM_MEAN`.
# Per default, `GEOM_MEAN` is used.
# Weights for each target can also be specified as a list of floats in the arguments
# Per default, weights are equally distributed between all targets and are normalized internally.
# It is thus not necessary to handle normalization or scaling.


objective = DesirabilityObjective(
    targets=targets,
    weights=[20, 20, 60],
    scalarizer="MEAN",
)

print(objective)

### Creating and printing the campaign

campaign = Campaign(searchspace=searchspace, objective=objective)
print(campaign)

### Performing some iterations

# The following loop performs some recommendations and adds fake results.
# It also prints what happens to internal data.

N_ITERATIONS = 3

for kIter in range(N_ITERATIONS):
    rec = campaign.recommend(batch_size=3)
    add_fake_measurements(rec, campaign.targets)
    campaign.add_measurements(rec)
    desirability = campaign.objective.transform(campaign.measurements, allow_extra=True)

    print(f"\n\n#### ITERATION {kIter+1} ####")
    print("\nRecommended measurements with fake measured results:\n")
    print(rec)
    print("\nInternal measurement database with desirability values:\n")
    print(pd.concat([campaign.measurements, desirability], axis=1))

### Addendum: Description of `transformation` functions

# This function is used to transform target values to the interval `[0,1]` for `MAX`/`MIN` mode.
# An ascending or decreasing `LINEAR` function is used per default.
# This function maps input values in a specified interval [lower, upper] to the interval `[0,1]`.
# Outside the specified interval, the function remains constant, that is, 0 or 1.

# For the match mode, two functions are available `TRIANGULAR` and `BELL`.
# The `TRIANGULAR` function is 0 outside a specified interval and linearly increases to 1 from both
# interval ends, reaching the value 1 at the center of the interval.
# This function is used per default for MATCH mode.
# The `BELL` function is a Gaussian bell curve, specified through the boundary values of the sigma
# interval, reaching the maximum value of 1 at the interval center.



"""Strategies for basic types."""

from functools import partial

import hypothesis.strategies as st
import numpy as np

from baybe.utils.numerical import DTypeFloatNumpy

finite_floats = partial(
    st.floats,
    allow_infinity=False,
    allow_nan=False,
    width=32 if DTypeFloatNumpy == np.float32 else 64,
)
"""A strategy producing finite (i.e., non-nan and non-infinite) floats."""

positive_finite_floats = partial(finite_floats, min_value=0.0, exclude_min=True)
""""A strategy producing positive finite (i.e., non-nan and non-infinite) floats."""



"""Tests for augmentation utilities."""

import pandas as pd
import pytest
from pytest import param

from baybe.utils.augmentation import (
    df_apply_dependency_augmentation,
    df_apply_permutation_augmentation,
)


@pytest.mark.parametrize(
    ("content", "col_groups", "content_expected"),
    [
        param(  # 2 invariant cols and 1 unaffected col
            {
                "data": {
                    "A": [1, 1],
                    "B": [2, 2],
                    "C": ["x", "y"],
                },
                "index": [0, 1],
            },
            [["A"], ["B"]],
            {
                "data": {
                    "A": [1, 2, 1, 2],
                    "B": [2, 1, 2, 1],
                    "C": ["x", "x", "y", "y"],
                },
                "index": [0, 0, 1, 1],
            },
            id="2inv_1add",
        ),
        param(  # 2 invariant cols with identical row values
            {
                "data": {
                    "A": [1, 1],
                    "B": [2, 2],
                },
                "index": [0, 1],
            },
            [["A"], ["B"]],
            {
                "data": {
                    "A": [1, 2, 1, 2],
                    "B": [2, 1, 2, 1],
                },
                "index": [0, 0, 1, 1],
            },
            id="2inv+degen",
        ),
        param(  # 2 groups with identical row values but different targets
            {
                "data": {
                    "A": [1, 1],
                    "B": [2, 2],
                    "T": ["x", "y"],
                },
                "index": [0, 1],
            },
            [["A"], ["B"]],
            {
                "data": {
                    "A": [1, 2, 1, 2],
                    "B": [2, 1, 2, 1],
                    "T": ["x", "x", "y", "y"],
                },
                "index": [0, 0, 1, 1],
            },
            id="2inv+degen_target",
        ),
        param(  # 2 groups with identical row values but same targets
            {
                "data": {
                    "A": [1, 1],
                    "B": [2, 2],
                    "T": ["x", "x"],
                },
                "index": [0, 1],
            },
            [["A"], ["B"]],
            {
                "data": {
                    "A": [1, 2, 1, 2],
                    "B": [2, 1, 2, 1],
                    "T": ["x", "x", "x", "x"],
                },
                "index": [0, 0, 1, 1],
            },
            id="2inv+degen_target+degen",
        ),
        param(  # 3 invariant groups with 1 entry each
            {
                "data": {
                    "A": [1, 1],
                    "B": [2, 4],
                    "C": [3, 5],
                    "D": ["x", "y"],
                },
                "index": [0, 1],
            },
            [["A"], ["B"], ["C"]],
            {
                "data": {
                    "A": [1, 1, 2, 2, 3, 3, 1, 1, 4, 4, 5, 5],
                    "B": [2, 3, 1, 3, 1, 2, 4, 5, 1, 5, 1, 4],
                    "C": [3, 2, 3, 1, 2, 1, 5, 4, 5, 1, 4, 1],
                    "D": ["x", "x", "x", "x", "x", "x", "y", "y", "y", "y", "y", "y"],
                },
                "index": [0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1],
            },
            id="3inv_1add",
        ),
        param(  # 2 groups with 2 entries each, 2 additional columns
            {
                "data": {
                    "Slot1": ["s1", "s2"],
                    "Slot2": ["s2", "s4"],
                    "Frac1": [0.1, 0.6],
                    "Frac2": [0.9, 0.4],
                    "Other1": ["A", "B"],
                    "Other2": ["C", "D"],
                },
                "index": [0, 1],
            },
            [["Slot1", "Frac1"], ["Slot2", "Frac2"]],
            {
                "data": {
                    "Slot1": ["s1", "s2", "s2", "s4"],
                    "Slot2": ["s2", "s1", "s4", "s2"],
                    "Frac1": [0.1, 0.9, 0.6, 0.4],
                    "Frac2": [0.9, 0.1, 0.4, 0.6],
                    "Other1": ["A", "A", "B", "B"],
                    "Other2": ["C", "C", "D", "D"],
                },
                "index": [0, 0, 1, 1],
            },
            id="2inv_2dependent_2add",
        ),
        param(  # 2 groups with 3 entries each, 1 additional column
            {
                "data": {
                    "Slot1": ["s1", "s2"],
                    "Slot2": ["s2", "s4"],
                    "Frac1": [0.1, 0.6],
                    "Frac2": [0.9, 0.4],
                    "Temp1": [10, 20],
                    "Temp2": [50, 60],
                    "Other": ["x", "y"],
                },
                "index": [0, 1],
            },
            [["Slot1", "Frac1", "Temp1"], ["Slot2", "Frac2", "Temp2"]],
            {
                "data": {
                    "Slot1": ["s1", "s2", "s2", "s4"],
                    "Slot2": ["s2", "s1", "s4", "s2"],
                    "Frac1": [0.1, 0.9, 0.6, 0.4],
                    "Frac2": [0.9, 0.1, 0.4, 0.6],
                    "Temp1": [10, 50, 20, 60],
                    "Temp2": [50, 10, 60, 20],
                    "Other": ["x", "x", "y", "y"],
                },
                "index": [0, 0, 1, 1],
            },
            id="2inv_4dependent2each_1add",
        ),
    ],
)
def test_df_permutation_aug(content, col_groups, content_expected):
    """Test permutation invariance data augmentation is done correctly."""
    # Create all needed dataframes
    df = pd.DataFrame(**content)
    df_augmented = df_apply_permutation_augmentation(df, col_groups)
    df_expected = pd.DataFrame(**content_expected)

    # Determine equality ignoring row order
    are_equal = df_augmented.equals(df_expected)

    assert (
        are_equal
    ), f"\norig:\n{df}\n\naugmented:\n{df_augmented}\n\nexpected:\n{df_expected}"


@pytest.mark.parametrize(
    ("col_groups", "msg"),
    [
        param([], "at least two column sequences", id="no_groups"),
        param([["A"]], "at least two column sequences", id="just_one_group"),
        param([["A"], ["B", "C"]], "the amount of columns in", id="different_lengths"),
        param([[], []], "each column group has", id="empty_group"),
    ],
)
def test_df_permutation_aug_invalid(col_groups, msg):
    """Test correct errors for invalid permutation attempts."""
    df = pd.DataFrame({"A": [1, 1], "B": [2, 2], "C": ["x", "y"]})
    with pytest.raises(ValueError, match=msg):
        df_apply_permutation_augmentation(df, col_groups)


@pytest.mark.parametrize(
    ("content", "causing", "affected", "content_expected"),
    [
        param(  # 1 causing val, 1 col affected (with 3 values)
            {
                "data": {
                    "A": [0, 1],
                    "B": [3, 4],
                    "C": ["x", "y"],
                },
                "index": [0, 1],
            },
            ("A", [0]),
            [("B", [3, 4, 5])],
            {
                "data": {
                    "A": [0, 0, 0, 1],
                    "B": [3, 4, 5, 4],
                    "C": ["x", "x", "x", "y"],
                },
                "index": [0, 0, 0, 1],
            },
            id="1causing_1affected",
        ),
        param(  # 1 causing val, 2 cols affected (with 2 values each)
            {
                "data": {
                    "A": [0, 1],
                    "B": [3, 4],
                    "C": ["x", "y"],
                },
                "index": [0, 1],
            },
            ("A", [0]),
            [("B", [3, 4]), ("C", ["x", "y"])],
            {
                "data": {
                    "A": [0, 0, 0, 0, 1],
                    "B": [3, 3, 4, 4, 4],
                    "C": ["x", "y", "x", "y", "y"],
                },
                "index": [0, 0, 0, 0, 1],
            },
            id="1causing_2affected",
        ),
        param(  # 2 causing vals, 1 col affected (with 3 values)
            {
                "data": {
                    "A": [0, 1, 2],
                    "B": [3, 4, 3],
                    "C": ["x", "y", "z"],
                },
                "index": [0, 1, 2],
            },
            ("A", [0, 1]),
            [("B", [3, 4, 5])],
            {
                "data": {
                    "A": [0, 0, 0, 1, 1, 1, 2],
                    "B": [3, 4, 5, 3, 4, 5, 3],
                    "C": ["x", "x", "x", "y", "y", "y", "z"],
                },
                "index": [0, 0, 0, 1, 1, 1, 2],
            },
            id="2causing_1affected",
        ),
        param(  # 2 causing vals, 2 cols affected (with 2 values each)
            {
                "data": {
                    "A": [0, 1, 2],
                    "B": [3, 4, 3],
                    "C": ["x", "y", "x"],
                },
                "index": [0, 1, 2],
            },
            ("A", [0, 1]),
            [("B", [3, 4]), ("C", ["x", "y"])],
            {
                "data": {
                    "A": [0, 0, 0, 0, 1, 1, 1, 1, 2],
                    "B": [3, 3, 4, 4, 3, 3, 4, 4, 3],
                    "C": ["x", "y", "x", "y", "x", "y", "x", "y", "x"],
                },
                "index": [0, 0, 0, 0, 1, 1, 1, 1, 2],
            },
            id="2causing_2affected",
        ),
    ],
)
def test_df_dependency_aug(content, causing, affected, content_expected):
    """Test dependency data augmentation is done correctly."""
    # Create all needed dataframes
    df = pd.DataFrame(**content)
    df_augmented = df_apply_dependency_augmentation(df, causing, affected)
    df_expected = pd.DataFrame(**content_expected)

    # Determine equality ignoring row order
    are_equal = df_augmented.equals(df_expected)

    assert (
        are_equal
    ), f"\norig:\n{df}\n\naugmented:\n{df_augmented}\n\nexpected:\n{df_expected}"



"""Validation functionality for search spaces."""

import warnings
from collections.abc import Collection, Sequence
from typing import TypeVar

import pandas as pd

from baybe.exceptions import EmptySearchSpaceError
from baybe.parameters import TaskParameter
from baybe.parameters.base import Parameter
from baybe.utils.dataframe import get_transform_objects

_T = TypeVar("_T", bound=Parameter)


def validate_parameter_names(  # noqa: DOC101, DOC103
    parameters: Collection[Parameter],
) -> None:
    """Validate the parameter names.

    Raises:
        ValueError: If the given list contains parameters with the same name.
    """
    param_names = [p.name for p in parameters]
    if len(set(param_names)) != len(param_names):
        raise ValueError("All parameters must have unique names.")


def validate_parameters(parameters: Collection[Parameter]) -> None:  # noqa: DOC101, DOC103
    """Validate the parameters.

    Raises:
        EmptySearchSpaceError: If the parameter list is empty.
        NotImplementedError: If more than one
            :class:`baybe.parameters.categorical.TaskParameter` is requested.
    """
    if not parameters:
        raise EmptySearchSpaceError("At least one parameter must be provided.")

    # TODO [16932]: Remove once more task parameters are supported
    if len([p for p in parameters if isinstance(p, TaskParameter)]) > 1:
        raise NotImplementedError(
            "Currently, at most one task parameter can be considered."
        )

    # Assert: unique names
    validate_parameter_names(parameters)


def get_transform_parameters(
    parameters: Sequence[_T],
    df: pd.DataFrame,
    allow_missing: bool = False,
    allow_extra: bool = False,
) -> list[_T]:
    """Deprecated!"""  # noqa: D401
    warnings.warn(
        f"The function 'get_transform_parameters' has been deprecated and will be "
        f"removed in a future version. Use '{get_transform_objects.__name__}' instead.",
        DeprecationWarning,
    )
    return get_transform_objects(
        df, parameters, allow_missing=allow_missing, allow_extra=allow_extra
    )



"""Scaling utilities."""

from __future__ import annotations

import gc
import itertools
from typing import TYPE_CHECKING

from attrs import define, field
from attrs.validators import deep_iterable, deep_mapping, instance_of

if TYPE_CHECKING:
    from botorch.models.transforms.input import InputTransform
    from torch import Tensor


@define
class ColumnTransformer:
    """Class for applying separate transforms to different column groups of tensors."""

    mapping: dict[tuple[int, ...], InputTransform] = field()
    """A mapping defining what transform to apply to which columns."""

    _is_trained: bool = field(default=False, init=False)
    """Boolean indicating if the transformer has been trained."""

    @mapping.validator
    def _validate_mapping_types_lazily(self, attr, value):
        """Perform transform ``isinstance`` check using lazy import."""
        from botorch.models.transforms.input import InputTransform

        validator = deep_mapping(
            mapping_validator=instance_of(dict),
            key_validator=deep_iterable(
                member_validator=instance_of(int), iterable_validator=instance_of(tuple)
            ),
            value_validator=instance_of(InputTransform),
        )
        validator(self, attr, value)

    @mapping.validator
    def _validate_mapping_is_disjoint(self, _, value: dict):
        """Validate that the each column is assigned to at most one transformer."""
        for x, y in itertools.combinations(value.keys(), 2):
            if not set(x).isdisjoint(y):
                raise ValueError(
                    f"The provided column specifications {x} and {y} are not disjoint."
                )

    def fit(self, x: Tensor, /) -> None:
        """Fit the transformer to the given tensor."""
        # Explicitly set flag to False to guarantee a clean state in case of
        # exceptions occurring in the for-loop below
        self._is_trained = False

        for cols, transformer in self.mapping.items():
            transformer.train()
            transformer(x[..., cols])
            transformer.eval()
        self._is_trained = True

    def transform(self, x: Tensor, /) -> Tensor:
        """Transform the given tensor."""
        if not self._is_trained:
            raise RuntimeError(
                f"The {self.__class__.__name__} must be trained before it can be used."
            )

        out = x.clone()
        for cols, transformer in self.mapping.items():
            out[..., cols] = transformer(out[..., cols])
        return out


# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



"""Tests for the custom parameter."""

import pytest

from .conftest import run_iterations


@pytest.mark.parametrize("parameter_names", [["Custom_1", "Custom_2"]])
def test_run_iterations(campaign, n_iterations, batch_size):
    """Test if iterative loop runs with custom parameters."""
    run_iterations(campaign, n_iterations, batch_size)



"""Base classes for all acquisition functions."""

from __future__ import annotations

import gc
import warnings
from abc import ABC
from inspect import signature
from typing import TYPE_CHECKING, ClassVar

import pandas as pd
from attrs import define

from baybe.exceptions import (
    IncompatibleAcquisitionFunctionError,
    UnidentifiedSubclassError,
)
from baybe.objectives.base import Objective
from baybe.objectives.desirability import DesirabilityObjective
from baybe.objectives.single import SingleTargetObjective
from baybe.searchspace.core import SearchSpace
from baybe.serialization.core import (
    converter,
    get_base_structure_hook,
    unstructure_base,
)
from baybe.serialization.mixin import SerialMixin
from baybe.surrogates.base import SurrogateProtocol
from baybe.targets.enum import TargetMode
from baybe.targets.numerical import NumericalTarget
from baybe.utils.basic import classproperty, match_attributes
from baybe.utils.boolean import is_abstract
from baybe.utils.dataframe import to_tensor

if TYPE_CHECKING:
    from botorch.acquisition import AcquisitionFunction as BotorchAcquisitionFunction


@define(frozen=True)
class AcquisitionFunction(ABC, SerialMixin):
    """Abstract base class for all acquisition functions."""

    abbreviation: ClassVar[str]
    """An alternative name for type resolution."""

    @classproperty
    def is_mc(cls) -> bool:
        """Flag indicating whether this is a Monte-Carlo acquisition function."""
        return cls.abbreviation.startswith("q")

    @classproperty
    def _non_botorch_attrs(cls) -> tuple[str, ...]:
        """Names of attributes that are not passed to the BoTorch constructor."""
        return ()

    def to_botorch(
        self,
        surrogate: SurrogateProtocol,
        searchspace: SearchSpace,
        objective: Objective,
        measurements: pd.DataFrame,
        pending_experiments: pd.DataFrame | None = None,
    ):
        """Create the botorch-ready representation of the function.

        The required structure of `measurements` is specified in
        :meth:`baybe.recommenders.base.RecommenderProtocol.recommend`.
        """
        import botorch.acquisition as bo_acqf
        import torch
        from botorch.acquisition.objective import LinearMCObjective

        from baybe.acquisition.acqfs import qThompsonSampling

        # Retrieve botorch acquisition function class and match attributes
        acqf_cls = _get_botorch_acqf_class(type(self))
        params_dict = match_attributes(
            self, acqf_cls.__init__, ignore=self._non_botorch_attrs
        )[0]

        # Create botorch surrogate model
        bo_surrogate = surrogate.to_botorch()

        # Get computational data representation
        train_x = to_tensor(searchspace.transform(measurements, allow_extra=True))

        # Collect remaining (context-specific) parameters
        signature_params = signature(acqf_cls).parameters
        additional_params = {}
        additional_params["model"] = bo_surrogate
        if "X_baseline" in signature_params:
            additional_params["X_baseline"] = train_x
        if "mc_points" in signature_params:
            additional_params["mc_points"] = to_tensor(
                self.get_integration_points(searchspace)  # type: ignore[attr-defined]
            )
        if pending_experiments is not None:
            if self.is_mc:
                pending_x = searchspace.transform(pending_experiments, allow_extra=True)
                additional_params["X_pending"] = to_tensor(pending_x)
            else:
                raise IncompatibleAcquisitionFunctionError(
                    f"Pending experiments were provided but the chosen acquisition "
                    f"function '{self.__class__.__name__}' does not support this."
                )

        # Add acquisition objective / best observed value
        match objective:
            case SingleTargetObjective(NumericalTarget(mode=TargetMode.MIN)):
                if "best_f" in signature_params:
                    additional_params["best_f"] = (
                        bo_surrogate.posterior(train_x).mean.min().item()
                    )
                    if issubclass(acqf_cls, bo_acqf.MCAcquisitionFunction):
                        additional_params["best_f"] *= -1.0

                if issubclass(acqf_cls, bo_acqf.AnalyticAcquisitionFunction):
                    additional_params["maximize"] = False
                elif issubclass(acqf_cls, bo_acqf.qNegIntegratedPosteriorVariance):
                    # qNIPV is valid but does not require any adjusted params
                    pass
                elif issubclass(acqf_cls, bo_acqf.MCAcquisitionFunction):
                    additional_params["objective"] = LinearMCObjective(
                        torch.tensor([-1.0])
                    )
                else:
                    raise ValueError(
                        f"Unsupported acquisition function type: {acqf_cls}."
                    )
            case SingleTargetObjective() | DesirabilityObjective():
                if "best_f" in signature_params:
                    additional_params["best_f"] = (
                        bo_surrogate.posterior(train_x).mean.max().item()
                    )
            case _:
                raise ValueError(f"Unsupported objective type: {objective}")

        params_dict.update(additional_params)

        acqf = acqf_cls(**params_dict)

        if isinstance(self, qThompsonSampling):
            assert hasattr(acqf, "_default_sample_shape")
            acqf._default_sample_shape = torch.Size([self.n_mc_samples])

        return acqf


def _get_botorch_acqf_class(
    baybe_acqf_cls: type[AcquisitionFunction], /
) -> type[BotorchAcquisitionFunction]:
    """Extract the BoTorch acquisition class for the given BayBE acquisition class."""
    import botorch

    for cls in baybe_acqf_cls.mro():
        if acqf_cls := getattr(botorch.acquisition, cls.__name__, False):
            if is_abstract(acqf_cls):
                continue
            return acqf_cls  # type: ignore

    raise UnidentifiedSubclassError(
        f"No BoTorch acquisition function class match found for "
        f"'{baybe_acqf_cls.__name__}'."
    )


# Register (un-)structure hooks
def _add_deprecation_hook(hook):
    """Add deprecation warnings to the default hook.

    Used for backward compatibility only and will be removed in future versions.
    """

    def added_deprecation_hook(val: dict | str, cls: type):
        # Backwards-compatibility needs to be ensured only for deserialization from
        # base class using string-based type specifiers as listed below,
        # since the concrete classes were available only after the change.
        if is_abstract(cls):
            UCB_DEPRECATIONS = {
                "VarUCB": "UpperConfidenceBound",
                "qVarUCB": "qUpperConfidenceBound",
            }
            if (
                entry := val if isinstance(val, str) else val["type"]
            ) in UCB_DEPRECATIONS:
                warnings.warn(
                    f"The use of `{entry}` is deprecated and will be disabled in a "
                    f"future version. To get the same outcome, use the new "
                    f"`{UCB_DEPRECATIONS[entry]}` class instead with a beta of 100.0.",
                    DeprecationWarning,
                )
                val = {"type": UCB_DEPRECATIONS[entry], "beta": 100.0}

        return hook(val, cls)

    return added_deprecation_hook


converter.register_structure_hook(
    AcquisitionFunction,
    _add_deprecation_hook(get_base_structure_hook(AcquisitionFunction)),
)
converter.register_unstructure_hook(AcquisitionFunction, unstructure_base)

# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



"""Tests pending experiments mechanism."""

import warnings

import pandas as pd
import pytest
from pytest import param

from baybe.acquisition.base import AcquisitionFunction
from baybe.exceptions import IncompatibleAcquisitionFunctionError, UnusedObjectWarning
from baybe.recommenders import (
    BotorchRecommender,
    FPSRecommender,
    GaussianMixtureClusteringRecommender,
    KMeansClusteringRecommender,
    NaiveHybridSpaceRecommender,
    PAMClusteringRecommender,
    TwoPhaseMetaRecommender,
)
from baybe.searchspace.core import SearchSpaceType
from baybe.utils.basic import get_subclasses
from baybe.utils.dataframe import add_fake_measurements, add_parameter_noise
from baybe.utils.random import temporary_seed

_discrete_params = ["Categorical_1", "Switch_1", "Num_disc_1"]
_continuous_params = ["Conti_finite1", "Conti_finite2", "Conti_finite3"]
_hybrid_params = ["Categorical_1", "Num_disc_1", "Conti_finite1", "Conti_finite2"]


@pytest.mark.parametrize(
    "parameter_names, recommender",
    [
        param(
            _discrete_params,
            FPSRecommender(),
            id="fps_discrete",
        ),
        param(_discrete_params, PAMClusteringRecommender(), id="pam_discrete"),
        param(
            _discrete_params,
            KMeansClusteringRecommender(),
            id="kmeans_discrete",
        ),
        param(
            _discrete_params,
            GaussianMixtureClusteringRecommender(),
            id="gm_discrete",
        ),
        param(
            _discrete_params,
            TwoPhaseMetaRecommender(recommender=BotorchRecommender()),
            id="botorch_discrete",
        ),
        param(
            _continuous_params,
            TwoPhaseMetaRecommender(recommender=BotorchRecommender()),
            id="botorch_continuous",
        ),
        param(
            _hybrid_params,
            TwoPhaseMetaRecommender(recommender=BotorchRecommender()),
            id="botorch_hybrid",
        ),
        param(
            _discrete_params,
            TwoPhaseMetaRecommender(recommender=BotorchRecommender()),
            id="botorch_discrete_allow",
        ),
        param(
            _continuous_params,
            TwoPhaseMetaRecommender(recommender=BotorchRecommender()),
            id="botorch_continuous_allow",
        ),
        param(
            _hybrid_params,
            TwoPhaseMetaRecommender(recommender=BotorchRecommender()),
            id="botorch_hybrid_allow",
        ),
        param(
            _discrete_params,
            NaiveHybridSpaceRecommender(
                disc_recommender=FPSRecommender(),
            ),
            id="naive1_discrete",
        ),
        param(
            _discrete_params,
            NaiveHybridSpaceRecommender(
                disc_recommender=KMeansClusteringRecommender(),
            ),
            id="naive2_discrete",
        ),
    ],
)
@pytest.mark.parametrize("n_grid_points", [8], ids=["grid8"])
def test_pending_points(campaign, batch_size):
    """Test there is no recommendation overlap if pending experiments are specified."""
    warnings.filterwarnings("ignore", category=UnusedObjectWarning)

    # Repeated recommendations explicitly need to be allowed or the potential overlap
    # will be avoided trivially
    if campaign.searchspace.type == SearchSpaceType.DISCRETE:
        campaign.allow_recommending_already_recommended = True
        campaign.allow_recommending_already_measured = True

    # Perform a fake first iteration
    rec = campaign.recommend(batch_size)
    add_fake_measurements(rec, campaign.targets)
    campaign.add_measurements(rec)

    # Get recommendations and set them as pending experiments while getting another set
    # Fix the random seed for each recommend call to limit influence of randomness in
    # some recommenders which could also trivially avoid overlap
    with temporary_seed(1337):
        rec1 = campaign.recommend(batch_size)
    campaign._cached_recommendation = pd.DataFrame()  # ensure no recommendation cache
    with temporary_seed(1337):
        rec2 = campaign.recommend(batch_size=batch_size, pending_experiments=rec1)

    # Assert they have no overlap, round to avoid numerical fluctuation
    overlap = pd.merge(rec1.round(3), rec2.round(3), how="inner")
    assert len(overlap) == 0, (
        f"Recommendations are overlapping!\n\nRecommendations 1:\n{rec1}\n\n"
        f"Recommendations 2:\n{rec2}\n\nOverlap:\n{overlap}"
    )


_non_mc_acqfs = [a() for a in get_subclasses(AcquisitionFunction) if not a.is_mc]


@pytest.mark.parametrize(
    "acqf", _non_mc_acqfs, ids=[a.abbreviation for a in _non_mc_acqfs]
)
@pytest.mark.parametrize(
    "parameter_names",
    [
        param(_discrete_params, id="discrete"),
        param(_continuous_params, id="continuous"),
        param(_hybrid_params, id="hybrid"),
    ],
)
@pytest.mark.parametrize("n_grid_points", [5], ids=["g5"])
@pytest.mark.parametrize("batch_size", [3], ids=["b3"])
def test_invalid_acqf(searchspace, recommender, objective, batch_size, acqf):
    """Test exception raised for acqfs that don't support pending experiments."""
    recommender = TwoPhaseMetaRecommender(
        recommender=BotorchRecommender(acquisition_function=acqf)
    )

    # Get recommendation and add a fake results
    rec1 = recommender.recommend(batch_size, searchspace, objective)
    add_fake_measurements(rec1, objective.targets)

    # Create fake pending experiments
    rec2 = rec1.copy()
    add_parameter_noise(rec2, searchspace.parameters)

    with pytest.raises(IncompatibleAcquisitionFunctionError):
        recommender.recommend(
            batch_size,
            searchspace,
            objective,
            measurements=rec1,
            pending_experiments=rec2,
        )



"""Functionality for single-target objectives."""

import gc
import warnings

import pandas as pd
from attrs import define, field
from attrs.validators import instance_of
from typing_extensions import override

from baybe.objectives.base import Objective
from baybe.targets.base import Target
from baybe.targets.enum import TargetMode
from baybe.targets.numerical import NumericalTarget
from baybe.utils.dataframe import get_transform_objects, pretty_print_df
from baybe.utils.plotting import to_string


@define(frozen=True, slots=False)
class SingleTargetObjective(Objective):
    """An objective focusing on a single target."""

    _target: Target = field(validator=instance_of(Target), alias="target")
    """The single target considered by the objective."""

    @override
    def __str__(self) -> str:
        targets_list = [target.summary() for target in self.targets]
        targets_df = pd.DataFrame(targets_list)

        fields = [
            to_string("Type", self.__class__.__name__, single_line=True),
            to_string("Targets", pretty_print_df(targets_df)),
        ]

        return to_string("Objective", *fields)

    @override
    @property
    def targets(self) -> tuple[Target, ...]:
        return (self._target,)

    @override
    def transform(
        self,
        df: pd.DataFrame | None = None,
        /,
        *,
        allow_missing: bool = False,
        allow_extra: bool | None = None,
        data: pd.DataFrame | None = None,
    ) -> pd.DataFrame:
        # >>>>>>>>>> Deprecation
        if not ((df is None) ^ (data is None)):
            raise ValueError(
                "Provide the dataframe to be transformed as argument to `df`."
            )

        if data is not None:
            df = data
            warnings.warn(
                "Providing the dataframe via the `data` argument is deprecated and "
                "will be removed in a future version. Please pass your dataframe "
                "as positional argument instead.",
                DeprecationWarning,
            )

        # Mypy does not infer from the above that `df` must be a dataframe here
        assert isinstance(df, pd.DataFrame)

        if allow_extra is None:
            allow_extra = True
            if set(df.columns) - {p.name for p in self.targets}:
                warnings.warn(
                    "For backward compatibility, the new `allow_extra` flag is set "
                    "to `True` when left unspecified. However, this behavior will be "
                    "changed in a future version. If you want to invoke the old "
                    "behavior, please explicitly set `allow_extra=True`.",
                    DeprecationWarning,
                )
        # <<<<<<<<<< Deprecation

        # Even for a single target, it is convenient to use the existing validation
        # machinery instead of re-implementing it
        get_transform_objects(
            df, [self._target], allow_missing=allow_missing, allow_extra=allow_extra
        )

        target_data = df[self._target.name].copy()

        out = self._target.transform(target_data).to_frame()

        # TODO: Remove hotfix (https://github.com/emdgroup/baybe/issues/460)
        if (
            isinstance(t := self._target, NumericalTarget)
            and t.mode is TargetMode.MIN
            and t.bounds.is_bounded
        ):
            out = -out
        return out


# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



"""Recommenders based on sampling."""

from typing import ClassVar

import numpy as np
import pandas as pd
from sklearn.preprocessing import StandardScaler
from typing_extensions import override

from baybe.recommenders.pure.nonpredictive.base import NonPredictiveRecommender
from baybe.searchspace import SearchSpace, SearchSpaceType, SubspaceDiscrete
from baybe.utils.plotting import to_string
from baybe.utils.sampling_algorithms import farthest_point_sampling


class RandomRecommender(NonPredictiveRecommender):
    """Recommends experiments randomly."""

    # Class variables
    compatibility: ClassVar[SearchSpaceType] = SearchSpaceType.HYBRID
    # See base class.

    @override
    def _recommend_hybrid(
        self,
        searchspace: SearchSpace,
        candidates_exp: pd.DataFrame,
        batch_size: int,
    ) -> pd.DataFrame:
        if searchspace.type == SearchSpaceType.DISCRETE:
            return candidates_exp.sample(batch_size)

        cont_random = searchspace.continuous.sample_uniform(batch_size=batch_size)
        if searchspace.type == SearchSpaceType.CONTINUOUS:
            return cont_random

        disc_candidates, _ = searchspace.discrete.get_candidates()

        # TODO decide mechanism if number of possible discrete candidates is smaller
        #  than batch size
        disc_random = disc_candidates.sample(
            n=batch_size,
            replace=len(disc_candidates) < batch_size,
        )

        cont_random.index = disc_random.index
        return pd.concat([disc_random, cont_random], axis=1)

    @override
    def __str__(self) -> str:
        fields = [to_string("Compatibility", self.compatibility, single_line=True)]
        return to_string(self.__class__.__name__, *fields)


class FPSRecommender(NonPredictiveRecommender):
    """An initial recommender that selects candidates via Farthest Point Sampling."""

    # Class variables
    compatibility: ClassVar[SearchSpaceType] = SearchSpaceType.DISCRETE
    # See base class.

    @override
    def _recommend_discrete(
        self,
        subspace_discrete: SubspaceDiscrete,
        candidates_exp: pd.DataFrame,
        batch_size: int,
    ) -> pd.Index:
        # Fit scaler on entire search space
        # TODO [Scaling]: scaling should be handled by search space object
        scaler = StandardScaler()
        scaler.fit(subspace_discrete.comp_rep)

        # Scale and sample
        candidates_comp = subspace_discrete.transform(candidates_exp)
        candidates_scaled = np.ascontiguousarray(scaler.transform(candidates_comp))
        ilocs = farthest_point_sampling(candidates_scaled, batch_size)
        return candidates_comp.index[ilocs]

    @override
    def __str__(self) -> str:
        fields = [to_string("Compatibility", self.compatibility, single_line=True)]
        return to_string(self.__class__.__name__, *fields)



"""Test for continuous parameters."""

import numpy as np
import pytest


@pytest.mark.parametrize(
    "parameter_names",
    [
        ["Conti_finite1"],
        ["Conti_finite2"],
        ["Conti_finite1", "Conti_finite2"],
    ],
)
def test_valid_configs(campaign):
    """Test whether the given settings work without error."""
    print(campaign.searchspace.continuous.comp_rep_bounds.values.flatten())

    assert all(
        np.issubdtype(type(itm), np.floating)
        for itm in campaign.searchspace.continuous.comp_rep_bounds.values.flatten()
    )



"""Optional import for the insights subpackage."""

from baybe.exceptions import OptionalImportError

try:
    import shap
except ModuleNotFoundError as ex:
    raise OptionalImportError(
        "Explainer functionality is unavailable because 'insights' is not installed."
        " Consider installing BayBE with 'insights' dependency, e.g. via "
        "`pip install baybe[insights]`."
    ) from ex

__all__ = [
    "shap",
]



"""Tests for insights subpackage."""

import inspect
from contextlib import nullcontext
from unittest import mock

import numpy as np
import pandas as pd
import pytest
from pytest import mark

from baybe._optional.info import INSIGHTS_INSTALLED
from baybe.exceptions import IncompatibleExplainerError

if not INSIGHTS_INSTALLED:
    pytest.skip("Optional insights package not installed.", allow_module_level=True)


from baybe import insights
from baybe._optional.insights import shap
from baybe.campaign import Campaign
from baybe.insights.shap import (
    NON_SHAP_EXPLAINERS,
    SHAP_EXPLAINERS,
    SHAP_PLOTS,
    KernelExplainer,
    SHAPInsight,
    _get_explainer_cls,
)
from tests.conftest import run_iterations


@pytest.fixture
def n_grid_points():
    return 5


@pytest.fixture
def n_iterations():
    return 2


@pytest.fixture
def batch_size():
    return 2


@pytest.fixture(
    params=[
        ["Conti_finite1", "Conti_finite2"],
        ["Num_disc_1", "Fraction_1"],
        ["Num_disc_1", "Conti_finite1"],
        ["Num_disc_1", "Categorical_1"],
        ["Conti_finite1", "Categorical_1"],
    ],
    ids=[
        "params_conti",
        "params_disc_num",
        "params_hybrid_num",
        "params_disc_cat",
        "params_hybrid_cat",
    ],
)
def parameter_names(request):
    return request.param


def _has_required_init_parameters(cls: type[shap.Explainer]) -> bool:
    """Check if non-shap initializer has required standard parameters."""
    REQUIRED_PARAMETERS = ["self", "model", "data"]
    init_signature = inspect.signature(cls.__init__)
    parameters = list(init_signature.parameters.keys())
    return parameters[:3] == REQUIRED_PARAMETERS


@pytest.mark.parametrize("explainer_name", NON_SHAP_EXPLAINERS)
def test_non_shap_signature(explainer_name):
    """Non-SHAP explainers must have the required signature."""
    assert _has_required_init_parameters(_get_explainer_cls(explainer_name))


def _test_shap_insight(campaign, explainer_cls, use_comp_rep, is_shap):
    """Helper function for general SHAP explainer tests."""
    context = nullcontext()
    if (
        (not use_comp_rep)
        and (explainer_cls != "KernelExplainer")
        and any(not p.is_numerical for p in campaign.parameters)
    ):
        # We expect a validation error in case an explanation with an unsupported
        # explainer type is attempted on a search space representation with
        # non-numerical entries
        context = pytest.raises(IncompatibleExplainerError)

    with context:
        shap_insight = SHAPInsight.from_campaign(
            campaign,
            explainer_cls=explainer_cls,
            use_comp_rep=use_comp_rep,
        )

        # Sanity check explainer
        assert isinstance(shap_insight, insights.SHAPInsight)
        assert isinstance(shap_insight.explainer, _get_explainer_cls(explainer_cls))
        assert shap_insight.uses_shap_explainer == is_shap

        # Sanity check explanation
        df = campaign.measurements[[p.name for p in campaign.parameters]]
        if use_comp_rep:
            df = campaign.searchspace.transform(df)
        shap_explanation = shap_insight.explain(df)
        assert isinstance(shap_explanation, shap.Explanation)


@mark.slow
@mark.parametrize("explainer_cls", SHAP_EXPLAINERS)
@mark.parametrize("use_comp_rep", [False, True], ids=["exp", "comp"])
def test_shap_explainers(ongoing_campaign, explainer_cls, use_comp_rep):
    """Test the explain functionalities with measurements."""
    _test_shap_insight(ongoing_campaign, explainer_cls, use_comp_rep, is_shap=True)


@mark.parametrize("explainer_cls", NON_SHAP_EXPLAINERS)
def test_non_shap_explainers(ongoing_campaign, explainer_cls):
    """Test the non-SHAP explainers in computational representation."""
    _test_shap_insight(
        ongoing_campaign, explainer_cls, use_comp_rep=True, is_shap=False
    )


@mark.slow
@mark.parametrize("explainer_cls", ["KernelExplainer"], ids=["KernelExplainer"])
@mark.parametrize("use_comp_rep", [False, True], ids=["exp", "comp"])
def test_invalid_explained_data(ongoing_campaign, explainer_cls, use_comp_rep):
    """Test invalid explained data."""
    shap_insight = SHAPInsight.from_campaign(
        ongoing_campaign,
        explainer_cls=explainer_cls,
        use_comp_rep=use_comp_rep,
    )
    df = pd.DataFrame({"Num_disc_1": [0, 2]})
    with pytest.raises(
        ValueError,
        match="The provided dataframe must have the same column names as used by "
        "the explainer object.",
    ):
        shap_insight.explain(df)


@mark.slow
@mark.parametrize("use_comp_rep", [False, True], ids=["exp", "comp"])
@mark.parametrize("plot_type", SHAP_PLOTS)
def test_plots(ongoing_campaign: Campaign, use_comp_rep, plot_type):
    """Test the default SHAP plots."""
    shap_insight = SHAPInsight.from_campaign(
        ongoing_campaign,
        use_comp_rep=use_comp_rep,
    )
    df = ongoing_campaign.measurements[[p.name for p in ongoing_campaign.parameters]]
    if use_comp_rep:
        df = ongoing_campaign.searchspace.transform(df)
    with mock.patch("matplotlib.pyplot.show"):
        shap_insight.plot(plot_type, df)


def test_updated_campaign_explanations(campaign, n_iterations, batch_size):
    """Test explanations for campaigns with updated measurements."""
    with pytest.raises(
        ValueError,
        match="The campaign does not contain any measurements.",
    ):
        SHAPInsight.from_campaign(campaign)

    run_iterations(campaign, n_iterations=n_iterations, batch_size=batch_size)
    shap_insight = SHAPInsight.from_campaign(campaign)
    df = campaign.measurements[[p.name for p in campaign.parameters]]
    explanation_1 = shap_insight.explain(df)

    run_iterations(campaign, n_iterations=n_iterations, batch_size=batch_size)
    shap_insight = SHAPInsight.from_campaign(campaign)
    df = campaign.measurements[[p.name for p in campaign.parameters]]
    explanation_2 = shap_insight.explain(df)

    assert explanation_1 != explanation_2, "SHAP explanations should not be identical."


def test_creation_from_recommender(ongoing_campaign):
    """Test the creation of SHAP insights from a recommender."""
    recommender = ongoing_campaign.recommender.recommender
    shap_insight = SHAPInsight.from_recommender(
        recommender,
        ongoing_campaign.searchspace,
        ongoing_campaign.objective,
        ongoing_campaign.measurements,
    )
    assert isinstance(shap_insight, insights.SHAPInsight)


def test_column_permutation():
    """Explaining data with permuted columns gives permuted explanations."""
    N = 10

    # Create insights object and test data
    background_data = pd.DataFrame(np.random.random((N, 3)), columns=["x", "y", "z"])
    explainer = KernelExplainer(lambda x: x, background_data)
    insights = SHAPInsight(explainer, background_data)
    df = pd.DataFrame(np.random.random((N, 3)), columns=["x", "y", "z"])

    # Regular column order
    ex1 = insights.explain(df)

    # Permuted column order
    ex2 = insights.explain(df[["z", "x", "y"]])[:, [1, 2, 0]]

    assert np.array_equal(ex1.values, ex2.values)



"""Discrete constraints."""

from __future__ import annotations

import gc
from collections.abc import Callable
from functools import reduce
from typing import TYPE_CHECKING, Any, ClassVar, cast

import pandas as pd
from attrs import define, field
from attrs.validators import in_, min_len
from typing_extensions import override

from baybe.constraints.base import CardinalityConstraint, DiscreteConstraint
from baybe.constraints.conditions import (
    Condition,
    ThresholdCondition,
    _threshold_operators,
    _valid_logic_combiners,
)
from baybe.serialization import (
    block_deserialization_hook,
    block_serialization_hook,
    converter,
)
from baybe.utils.basic import Dummy

if TYPE_CHECKING:
    import polars as pl


@define
class DiscreteExcludeConstraint(DiscreteConstraint):
    """Class for modelling exclusion constraints."""

    # object variables
    conditions: list[Condition] = field(validator=min_len(1))
    """List of individual conditions."""

    combiner: str = field(default="AND", validator=in_(_valid_logic_combiners))
    """Operator encoding how to combine the individual conditions."""

    @override
    def get_invalid(self, data: pd.DataFrame) -> pd.Index:
        satisfied = [
            cond.evaluate(data[self.parameters[k]])
            for k, cond in enumerate(self.conditions)
        ]
        res = reduce(_valid_logic_combiners[self.combiner], satisfied)
        return data.index[res]

    @override
    def get_invalid_polars(self) -> pl.Expr:
        from baybe._optional.polars import polars as pl

        satisfied = []
        for k, cond in enumerate(self.conditions):
            satisfied.append(cond.to_polars(pl.col(self.parameters[k])))

        expr = pl.reduce(_valid_logic_combiners[self.combiner], satisfied)

        return expr


@define
class DiscreteSumConstraint(DiscreteConstraint):
    """Class for modelling sum constraints."""

    # IMPROVE: refactor `SumConstraint` and `ProdConstraint` to avoid code copying

    # class variables
    numerical_only: ClassVar[bool] = True
    # see base class.

    # object variables
    condition: ThresholdCondition = field()
    """The condition modeled by this constraint."""

    @override
    def get_invalid(self, data: pd.DataFrame) -> pd.Index:
        evaluate_data = data[self.parameters].sum(axis=1)
        mask_bad = ~self.condition.evaluate(evaluate_data)

        return data.index[mask_bad]

    @override
    def get_invalid_polars(self) -> pl.Expr:
        from baybe._optional.polars import polars as pl

        return self.condition.to_polars(pl.sum_horizontal(self.parameters)).not_()


@define
class DiscreteProductConstraint(DiscreteConstraint):
    """Class for modelling product constraints."""

    # IMPROVE: refactor `SumConstraint` and `ProdConstraint` to avoid code copying

    # class variables
    numerical_only: ClassVar[bool] = True
    # see base class.

    # object variables
    condition: ThresholdCondition = field()
    """The condition that is used for this constraint."""

    @override
    def get_invalid(self, data: pd.DataFrame) -> pd.Index:
        evaluate_data = data[self.parameters].prod(axis=1)
        mask_bad = ~self.condition.evaluate(evaluate_data)

        return data.index[mask_bad]

    @override
    def get_invalid_polars(self) -> pl.Expr:
        from baybe._optional.polars import polars as pl

        op = _threshold_operators[self.condition.operator]

        # Get the product of columns
        expr = pl.reduce(lambda acc, x: acc * x, pl.col(self.parameters))

        # Apply the threshold operator on expr and the condition threshold
        return op(expr, self.condition.threshold).not_()


class DiscreteNoLabelDuplicatesConstraint(DiscreteConstraint):
    """Constraint class for excluding entries where occurring labels are not unique.

    This can be useful to remove entries that arise from e.g. a permutation invariance
    as for instance here:

    - A,B,C,D would remain
    - A,A,B,C would be removed
    - A,A,B,B would be removed
    - A,A,B,A would be removed
    - A,C,A,C would be removed
    - A,C,B,C would be removed
    """

    @override
    def get_invalid(self, data: pd.DataFrame) -> pd.Index:
        mask_bad = data[self.parameters].nunique(axis=1) != len(self.parameters)

        return data.index[mask_bad]

    @override
    def get_invalid_polars(self) -> pl.Expr:
        from baybe._optional.polars import polars as pl

        expr = (
            pl.concat_list(pl.col(self.parameters))
            .list.eval(pl.element().n_unique())
            .explode()
        ) != len(self.parameters)

        return expr


class DiscreteLinkedParametersConstraint(DiscreteConstraint):
    """Constraint class for linking the values of parameters.

    This constraint type effectively allows generating parameter sets that relate to
    the same underlying quantity, e.g. two parameters that represent the same molecule
    using different encodings. Linking the parameters removes all entries from the
    search space where the parameter values differ.
    """

    @override
    def get_invalid(self, data: pd.DataFrame) -> pd.Index:
        mask_bad = data[self.parameters].nunique(axis=1) != 1

        return data.index[mask_bad]

    @override
    def get_invalid_polars(self) -> pl.Expr:
        from baybe._optional.polars import polars as pl

        expr = (
            pl.concat_list(pl.col(self.parameters))
            .list.eval(pl.element().n_unique())
            .explode()
        ) != 1

        return expr


@define
class DiscreteDependenciesConstraint(DiscreteConstraint):
    """Constraint that specifies dependencies between parameters.

    For instance some parameters might only be relevant when another parameter has a
    certain value (e.g. parameter switch is 'on'). All dependencies must be declared in
    a single constraint.
    """

    # class variables
    eval_during_augmentation: ClassVar[bool] = True
    # See base class

    # object variables
    conditions: list[Condition] = field()
    """The list of individual conditions."""

    affected_parameters: list[list[str]] = field()
    """The parameters affected by the individual conditions."""

    # for internal use only
    permutation_invariant: bool = field(default=False, init=False)
    """Flag that indicates whether the affected parameters are permutation invariant.
    This should not be changed by the user but by other constraints using the class."""

    @affected_parameters.validator
    def _validate_affected_parameters(  # noqa: DOC101, DOC103
        self, _: Any, value: list[list[str]]
    ) -> None:
        """Validate the affected parameters.

        Raises:
            ValueError: If one set of affected parameters does not have exactly one
                condition.
        """
        if len(self.conditions) != len(value):
            raise ValueError(
                f"For the {self.__class__.__name__}, for each item in the "
                f"affected_parameters list you must provide exactly one condition in "
                f"the conditions list."
            )

    @override
    def get_invalid(self, data: pd.DataFrame) -> pd.Index:
        # Create data copy and mark entries where the dependency conditions are negative
        # with a dummy value to cause degeneracy.
        censored_data = data.copy()
        for k, _ in enumerate(self.parameters):
            # .loc assignments are not supported by mypy + pandas-stubs yet
            # See https://github.com/pandas-dev/pandas-stubs/issues/572
            censored_data.loc[  # type: ignore[call-overload]
                ~self.conditions[k].evaluate(data[self.parameters[k]]),
                self.affected_parameters[k],
            ] = Dummy()

        # Create an invariant indicator: pair each value of an affected parameter with
        # the corresponding value of the parameter it depends on. These indicators
        # will become invariant when frozenset is applied to them.
        for k, param in enumerate(self.parameters):
            for affected_param in self.affected_parameters[k]:
                censored_data[affected_param] = list(
                    zip(censored_data[affected_param], censored_data[param])
                )

        # Merge the invariant indicator with all other parameters (i.e. neither the
        # affected nor the dependency-causing ones) and detect duplicates in that space.
        all_affected_params = [col for cols in self.affected_parameters for col in cols]
        other_params = (
            data.columns.drop(all_affected_params).drop(self.parameters).tolist()
        )
        df_eval = pd.concat(
            [
                censored_data[other_params],
                censored_data[all_affected_params].apply(
                    cast(Callable, frozenset)
                    if self.permutation_invariant
                    else cast(Callable, tuple),
                    axis=1,
                ),
            ],
            axis=1,
        )
        inds_bad = data.index[df_eval.duplicated(keep="first")]

        return inds_bad


@define
class DiscretePermutationInvarianceConstraint(DiscreteConstraint):
    """Constraint class for declaring that a set of parameters is permutation invariant.

    More precisely, this means that, ``(val_from_param1, val_from_param2)`` is
    equivalent to ``(val_from_param2, val_from_param1)``. Since it does not make sense
    to have this constraint with duplicated labels, this implementation also internally
    applies the :class:`baybe.constraints.discrete.DiscreteNoLabelDuplicatesConstraint`.

    *Note:* This constraint is evaluated during creation. In the future it might also be
    evaluated during modeling to make use of the invariance.
    """

    # class variables
    eval_during_augmentation: ClassVar[bool] = True
    # See base class

    # object variables
    dependencies: DiscreteDependenciesConstraint | None = field(default=None)
    """Dependencies connected with the invariant parameters."""

    @override
    def get_invalid(self, data: pd.DataFrame) -> pd.Index:
        # Get indices of entries with duplicate label entries. These will also be
        # dropped by this constraint.
        mask_duplicate_labels = pd.Series(False, index=data.index)
        mask_duplicate_labels[
            DiscreteNoLabelDuplicatesConstraint(parameters=self.parameters).get_invalid(
                data
            )
        ] = True

        # Merge a permutation invariant representation of all affected parameters with
        # the other parameters and indicate duplicates. This ensures that variation in
        # other parameters is also accounted for.
        other_params = data.columns.drop(self.parameters).tolist()
        df_eval = pd.concat(
            [
                data[other_params].copy(),
                data[self.parameters].apply(cast(Callable, frozenset), axis=1),
            ],
            axis=1,
        ).loc[
            ~mask_duplicate_labels  # only consider label-duplicate-free part
        ]
        mask_duplicate_permutations = df_eval.duplicated(keep="first")

        # Indices of entries with label-duplicates
        inds_duplicate_labels = data.index[mask_duplicate_labels]

        # Indices of duplicate permutations in the (already label-duplicate-free) data
        inds_duplicate_permutations = df_eval.index[mask_duplicate_permutations]

        # If there are dependencies connected to the invariant parameters evaluate them
        # here and remove resulting duplicates with a DependenciesConstraint
        inds_invalid = inds_duplicate_labels.union(inds_duplicate_permutations)
        if self.dependencies:
            self.dependencies.permutation_invariant = True
            inds_duplicate_independency_adjusted = self.dependencies.get_invalid(
                data.drop(index=inds_invalid)
            )
            inds_invalid = inds_invalid.union(inds_duplicate_independency_adjusted)

        return inds_invalid


@define
class DiscreteCustomConstraint(DiscreteConstraint):
    """Class for user-defined custom constraints."""

    # object variables
    validator: Callable[[pd.DataFrame], pd.Series] = field()
    """A user-defined function modeling the validation of the constraint. The expected
    return is a pandas series with Boolean entries True/False for search space elements
    you want to keep/remove."""

    @override
    def get_invalid(self, data: pd.DataFrame) -> pd.Index:
        mask_bad = ~self.validator(data[self.parameters])

        return data.index[mask_bad]


@define
class DiscreteCardinalityConstraint(CardinalityConstraint, DiscreteConstraint):
    """Class for discrete cardinality constraints."""

    # Class variables
    numerical_only: ClassVar[bool] = True
    # See base class.

    @override
    def get_invalid(self, data: pd.DataFrame) -> pd.Index:
        non_zeros = (data[self.parameters] != 0.0).sum(axis=1)
        mask_bad = non_zeros > self.max_cardinality
        mask_bad |= non_zeros < self.min_cardinality
        return data.index[mask_bad]


# Constraints are approximately ordered according to increasing computational effort
# to minimize total time in their sequential application
DISCRETE_CONSTRAINTS_FILTERING_ORDER = (
    DiscreteExcludeConstraint,
    DiscreteNoLabelDuplicatesConstraint,
    DiscreteLinkedParametersConstraint,
    DiscreteSumConstraint,
    DiscreteProductConstraint,
    DiscreteCardinalityConstraint,
    DiscreteCustomConstraint,
    DiscretePermutationInvarianceConstraint,
    DiscreteDependenciesConstraint,
)

# Prevent (de-)serialization of custom constraints
converter.register_unstructure_hook(DiscreteCustomConstraint, block_serialization_hook)
converter.register_structure_hook(DiscreteCustomConstraint, block_deserialization_hook)

# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()
